from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable

import PIL.Image

from sciencebeam_parser.utils.lazy import LazyLoaded, Preloadable


class OpticalCharacterRecognitionModelResult(ABC):
    @abstractmethod
    def get_text(self) -> str:
        pass


@dataclass
class SimpleOpticalCharacterRecognitionModelResult(OpticalCharacterRecognitionModelResult):
    text: str

    def get_text(self) -> str:
        return self.text


class OpticalCharacterRecognitionModel(ABC, Preloadable):
    @abstractmethod
    def predict_single(self, image: PIL.Image.Image) -> OpticalCharacterRecognitionModelResult:
        pass


T_OpticalCharacterRecognitionModelFactory = Callable[[], OpticalCharacterRecognitionModel]


class LazyOpticalCharacterRecognitionModel(OpticalCharacterRecognitionModel):
    def __init__(self, factory: T_OpticalCharacterRecognitionModelFactory) -> None:
        super().__init__()
        self._lazy_model = LazyLoaded[OpticalCharacterRecognitionModel](factory)

    def __repr__(self) -> str:
        return '%s(factory=%r, loaded=%r)' % (
            type(self).__name__, self._lazy_model.factory, self._lazy_model.is_loaded
        )

    @property
    def ocr_model(self) -> OpticalCharacterRecognitionModel:
        return self._lazy_model.get()

    def preload(self):
        self.ocr_model.preload()

    def predict_single(self, image: PIL.Image.Image) -> OpticalCharacterRecognitionModelResult:
        return self.ocr_model.predict_single(image)



import logging

from sciencebeam_parser.models.training_data import (
    NO_NS_TEI_E,
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/FullTextParser.java

ROOT_TRAINING_XML_ELEMENT_PATH = ['text']

TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<other>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="other"]'],
    '<section>': ROOT_TRAINING_XML_ELEMENT_PATH + ['head'],
    '<paragraph>': ROOT_TRAINING_XML_ELEMENT_PATH + ['p'],
    '<citation_marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['p', 'ref[@type="biblio"]'],
    '<figure_marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['p', 'ref[@type="figure"]'],
    '<table_marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['p', 'ref[@type="table"]'],
    '<equation_marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['p', 'ref[@type="formula"]'],
    '<section_marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['p', 'ref[@type="section"]'],
    '<figure>': ROOT_TRAINING_XML_ELEMENT_PATH + ['figure'],
    '<table>': ROOT_TRAINING_XML_ELEMENT_PATH + ['figure[@type="table"]'],
    '<equation>': ROOT_TRAINING_XML_ELEMENT_PATH + ['formula'],
    '<equation_label>': ROOT_TRAINING_XML_ELEMENT_PATH + ['formula', 'label'],
    '<item>': ROOT_TRAINING_XML_ELEMENT_PATH + ['item']
}


class FullTextTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.fulltext.tei.xml'
    DEFAULT_DATA_FILENAME_SUFFIX = '.fulltext'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            element_maker=NO_NS_TEI_E,
            default_tei_filename_suffix=(
                FullTextTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=(
                FullTextTeiTrainingDataGenerator.DEFAULT_DATA_FILENAME_SUFFIX
            ),
            default_tei_sub_directory='fulltext/corpus/tei',
            default_data_sub_directory='fulltext/corpus/raw'
        )


class FullTextTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL
            ),
            use_tei_namespace=False
        )



import logging
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union
)

from lxml import etree

from sciencebeam_parser.utils.xml_writer import TagExpression
from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.document.semantic_document import (
    SemanticAddressLine,
    SemanticAuthor,
    SemanticContentWrapper,
    SemanticCountry,
    SemanticDepartment,
    SemanticExternalIdentifier,
    SemanticExternalUrl,
    SemanticFigure,
    SemanticFigureCitation,
    SemanticGivenName,
    SemanticGraphic,
    SemanticHeading,
    SemanticInstitution,
    SemanticIssue,
    SemanticJournal,
    SemanticLabel,
    SemanticLaboratory,
    SemanticLocation,
    SemanticMarker,
    SemanticMiddleName,
    SemanticMixedNote,
    SemanticNameSuffix,
    SemanticNameTitle,
    SemanticNote,
    SemanticOptionalValueSemanticMixedContentWrapper,
    SemanticPageRange,
    SemanticParagraph,
    SemanticPostBox,
    SemanticPostCode,
    SemanticPublisher,
    SemanticRawEditors,
    SemanticRawEquation,
    SemanticReferenceCitation,
    SemanticRegion,
    SemanticSection,
    SemanticSettlement,
    SemanticSurname,
    SemanticTable,
    SemanticTableCitation,
    SemanticTextContentWrapper,
    SemanticTitle,
    SemanticVolume
)
from sciencebeam_parser.document.tei.common import (
    TEI_E,
    create_tei_note_element,
    extend_element,
    get_default_attributes_for_semantic_content,
    iter_layout_block_tei_children,
    parse_tag_expression
)
from sciencebeam_parser.document.tei.factory import (
    T_ElementChildrenList,
    TeiElementFactory,
    TeiElementFactoryContext
)
from sciencebeam_parser.document.tei.misc import (
    SemanticMixedNoteTeiElementFactory
)
from sciencebeam_parser.document.tei.citation import (
    CitationTeiElementFactory
)
from sciencebeam_parser.document.tei.external_identifiers import (
    ExternalIdentifierTeiElementFactory
)
from sciencebeam_parser.document.tei.equation import (
    RawEquationTeiElementFactory
)
from sciencebeam_parser.document.tei.figure_table import (
    FigureTeiElementFactory,
    TableTeiElementFactory
)
from sciencebeam_parser.document.tei.graphic import (
    GraphicTeiElementFactory
)
from sciencebeam_parser.document.tei.page_range import (
    TeiBiblScopeForPageRangeElementFactory
)
from sciencebeam_parser.document.tei.section import (
    HeadingTeiElementFactory,
    ParagraphTeiElementFactory,
    SectionTeiElementFactory
)


LOGGER = logging.getLogger(__name__)


SIMPLE_TAG_EXPRESSION_BY_SEMANTIC_CONTENT_CLASS = {
    SemanticNameTitle: 'roleName',
    SemanticGivenName: 'forename[@type="first"]',
    SemanticMiddleName: 'forename[@type="middle"]',
    SemanticSurname: 'surname',
    SemanticNameSuffix: 'genName',
    SemanticRawEditors: 'editor',
    SemanticLabel: 'label',
    SemanticMarker: 'note[@type="marker"]',
    SemanticInstitution: 'orgName[@type="institution"]',
    SemanticDepartment: 'orgName[@type="department"]',
    SemanticLaboratory: 'orgName[@type="laboratory"]',
    SemanticAddressLine: 'addrLine',
    SemanticPostCode: 'postCode',
    SemanticPostBox: 'postBox',
    SemanticSettlement: 'settlement',
    SemanticRegion: 'region',
    SemanticCountry: 'country',
    SemanticJournal: 'title[@level="j"]',
    SemanticVolume: 'biblScope[@unit="volume"]',
    SemanticIssue: 'biblScope[@unit="issue"]',
    SemanticPublisher: 'publisher',
    SemanticLocation: 'addrLine',
    SemanticExternalUrl: 'ref[@type="url"]'
}


PARSED_TAG_EXPRESSION_BY_SEMANTIC_CONTENT_CLASS: Dict[type, TagExpression] = {
    key: parse_tag_expression(value)
    for key, value in SIMPLE_TAG_EXPRESSION_BY_SEMANTIC_CONTENT_CLASS.items()
}


PARENT_PATH_BY_SEMANTIC_CONTENT_CLASS = {
    SemanticTitle: ['analytic'],
    SemanticAuthor: ['analytic'],
    SemanticRawEditors: ['monogr'],
    SemanticExternalIdentifier: ['analytic'],
    SemanticJournal: ['monogr'],
    SemanticVolume: ['monogr', 'imprint'],
    SemanticIssue: ['monogr', 'imprint'],
    SemanticPageRange: ['monogr', 'imprint'],
    SemanticPublisher: ['monogr', 'imprint'],
    SemanticLocation: ['monogr', 'meeting', 'address']
}


ELEMENT_FACTORY_BY_SEMANTIC_CONTENT_CLASS: Mapping[
    Type[Any],
    TeiElementFactory
] = {
    SemanticMixedNote: SemanticMixedNoteTeiElementFactory(),
    SemanticPageRange: TeiBiblScopeForPageRangeElementFactory(),
    SemanticExternalIdentifier: ExternalIdentifierTeiElementFactory(),
    SemanticFigure: FigureTeiElementFactory(),
    SemanticTable: TableTeiElementFactory(),
    SemanticFigureCitation: CitationTeiElementFactory(),
    SemanticTableCitation: CitationTeiElementFactory(),
    SemanticReferenceCitation: CitationTeiElementFactory(),
    SemanticHeading: HeadingTeiElementFactory(),
    SemanticParagraph: ParagraphTeiElementFactory(),
    SemanticSection: SectionTeiElementFactory(),
    SemanticRawEquation: RawEquationTeiElementFactory(),
    SemanticGraphic: GraphicTeiElementFactory()
}


def get_tei_note_for_semantic_content(
    semantic_content: SemanticContentWrapper
) -> etree.ElementBase:
    note_type = 'other'
    if isinstance(semantic_content, SemanticNote):
        note_type = semantic_content.note_type
    else:
        note_type = 'other:' + type(semantic_content).__name__
    return create_tei_note_element(note_type, semantic_content.merged_block)


class DefaultTeiElementFactoryContext(TeiElementFactoryContext):
    def get_default_attributes_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        **kwargs
    ) -> Dict[str, str]:
        return get_default_attributes_for_semantic_content(semantic_content, **kwargs)

    def iter_layout_block_tei_children(
        self,
        layout_block: LayoutBlock,
        enable_coordinates: bool = True
    ) -> Iterable[Union[str, etree.ElementBase]]:
        return iter_layout_block_tei_children(
            layout_block,
            enable_coordinates=enable_coordinates
        )

    def get_tei_child_elements_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper
    ) -> List[etree.ElementBase]:
        return get_tei_child_elements_for_semantic_content(
            semantic_content,
            context=self
        )

    def append_tei_children_list_and_get_whitespace(
        self,
        children: T_ElementChildrenList,
        semantic_content: SemanticContentWrapper,
        pending_whitespace: str
    ) -> str:
        return append_tei_children_list_and_get_whitespace(
            children,
            semantic_content=semantic_content,
            pending_whitespace=pending_whitespace,
            context=self
        )

    def get_parent_path_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper
    ) -> Optional[List[str]]:
        return PARENT_PATH_BY_SEMANTIC_CONTENT_CLASS.get(
            type(semantic_content)
        )


DEFAULT_TEI_ELEMENT_FACTORY_CONTEXT = DefaultTeiElementFactoryContext()


def get_tei_child_elements_for_semantic_content(
    semantic_content: SemanticContentWrapper,
    context: TeiElementFactoryContext
) -> List[etree.ElementBase]:
    semantic_type = type(semantic_content)
    element_factory = ELEMENT_FACTORY_BY_SEMANTIC_CONTENT_CLASS.get(
        semantic_type
    )
    if element_factory is not None:
        return element_factory.get_tei_children_for_semantic_content(
            semantic_content,
            context=context
        )
    parsed_tag_expression = PARSED_TAG_EXPRESSION_BY_SEMANTIC_CONTENT_CLASS.get(
        semantic_type
    )
    if parsed_tag_expression:
        if (
            isinstance(semantic_content, SemanticOptionalValueSemanticMixedContentWrapper)
            and semantic_content.value is not None
        ):
            return [parsed_tag_expression.create_node(
                get_default_attributes_for_semantic_content(semantic_content),
                semantic_content.value,
                element_maker=TEI_E
            )]
        return [parsed_tag_expression.create_node(
            *iter_layout_block_tei_children(semantic_content.merged_block),
            element_maker=TEI_E
        )]
    return [get_tei_note_for_semantic_content(semantic_content)]


def get_tei_children_and_whitespace_for_semantic_content(
    semantic_content: SemanticContentWrapper,
    context: TeiElementFactoryContext
) -> Tuple[List[Union[dict, str, etree.ElementBase]], str]:
    layout_block = semantic_content.merged_block
    if isinstance(semantic_content, SemanticTextContentWrapper):
        return (
            list(iter_layout_block_tei_children(layout_block)),
            layout_block.whitespace
        )
    return (
        get_tei_child_elements_for_semantic_content(
            semantic_content,
            context=context
        ),
        layout_block.whitespace
    )


def append_tei_children_and_get_whitespace(
    parent: etree.ElementBase,
    semantic_content: SemanticContentWrapper,
    context: TeiElementFactoryContext
) -> str:
    children, whitespace = get_tei_children_and_whitespace_for_semantic_content(
        semantic_content,
        context=context
    )
    extend_element(parent, children)
    return whitespace


def append_tei_children_list_and_get_whitespace(
    children: T_ElementChildrenList,
    semantic_content: SemanticContentWrapper,
    pending_whitespace: str,
    context: TeiElementFactoryContext
) -> str:
    tail_children, tail_whitespace = get_tei_children_and_whitespace_for_semantic_content(
        semantic_content,
        context=context
    )
    if not tail_children:
        return pending_whitespace
    if pending_whitespace:
        children.append(pending_whitespace)
    children.extend(tail_children)
    return tail_whitespace



import logging
from typing import Iterable, Mapping, Optional, Tuple

from sciencebeam_parser.utils.misc import iter_ids
from sciencebeam_parser.document.semantic_document import (
    SemanticAddressLine,
    SemanticAffiliationAddress,
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticCountry,
    SemanticDepartment,
    SemanticInstitution,
    SemanticLaboratory,
    SemanticMarker,
    SemanticNote,
    SemanticPostBox,
    SemanticPostCode,
    SemanticRegion,
    SemanticSettlement
)
from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.models.extract import (
    SimpleModelSemanticExtractor,
    get_regex_cleaned_layout_block_with_prefix_suffix
)


LOGGER = logging.getLogger(__name__)


SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<institution>': SemanticInstitution,
    '<department>': SemanticDepartment,
    '<laboratory>': SemanticLaboratory,
    '<addrLine>': SemanticAddressLine,
    '<postCode>': SemanticPostCode,
    '<postBox>': SemanticPostBox,
    '<region>': SemanticRegion,
    '<settlement>': SemanticSettlement,
    '<country>': SemanticCountry
}

CLEAN_REGEX_BY_TAG: Mapping[str, str] = {
    '<country>': r'(.*[^.]).*'
}


class AffiliationAddressSemanticExtractor(SimpleModelSemanticExtractor):
    def __init__(self):
        super().__init__(semantic_content_class_by_tag=SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG)

    def iter_semantic_content_for_entity_blocks(
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        ids_iterator = iter(iter_ids('aff'))
        aff: Optional[SemanticAffiliationAddress] = None
        for name, layout_block in entity_tokens:
            if name == '<marker>':
                if aff:
                    yield aff
                aff = SemanticAffiliationAddress(content_id=next(ids_iterator, '?'))
                aff.add_content(SemanticMarker(layout_block=layout_block))
                continue
            prefix_block, cleaned_block, suffix_block = (
                get_regex_cleaned_layout_block_with_prefix_suffix(
                    layout_block,
                    CLEAN_REGEX_BY_TAG.get(name)
                )
            )
            semantic_content = self.get_semantic_content_for_entity_name(
                name, cleaned_block
            )
            if (
                aff is not None
                and isinstance(semantic_content, SemanticInstitution)
                and aff.has_type(SemanticInstitution)
            ):
                yield aff
                aff = None
            if not aff:
                if isinstance(semantic_content, SemanticNote):
                    yield semantic_content
                    continue
                aff = SemanticAffiliationAddress(content_id=next(ids_iterator, '?'))
            if prefix_block:
                aff.add_content(SemanticNote(layout_block=prefix_block, note_type=f'{name}-prefix'))
            aff.add_content(semantic_content)
            if suffix_block:
                aff.add_content(SemanticNote(layout_block=suffix_block, note_type=f'{name}-suffix'))
        if aff:
            yield aff



import os
import re


def get_github_content_link_prefix(
    project_url: str,
    version: str,
    non_release_branch: str = 'develop'
) -> str:
    if 'dev' in version:
        return f'{project_url}/blob/{non_release_branch}'
    return f'{project_url}/blob/{version}'


def get_resolved_link(
    link: str,
    source_base_path: str,
    link_prefix: str
) -> str:
    result = os.path.join(link_prefix, source_base_path, link)
    if source_base_path and '/' not in source_base_path:
        result = result.replace(f'/{source_base_path}/../', '/')
    return result


def get_markdown_with_resolved_links(
    markdown_content: str,
    source_base_path: str,
    link_prefix: str
) -> str:
    return re.sub(
        r'(\[.*\]\()(.*)(\))',
        lambda m: (
            m.group(1)
            + get_resolved_link(m.group(2), source_base_path, link_prefix)
            + m.group(3)
        ),
        markdown_content
    )



from typing import Iterable

from sciencebeam_parser.models.data import (
    ContextAwareLayoutTokenFeatures,
    ContextAwareLayoutTokenModelDataGenerator,
    LayoutModelData
)


class AffiliationAddressDataGenerator(ContextAwareLayoutTokenModelDataGenerator):
    def iter_model_data_for_context_layout_token_features(
        self,
        token_features: ContextAwareLayoutTokenFeatures
    ) -> Iterable[LayoutModelData]:
        # using dummy line status due to https://github.com/kermitt2/grobid/issues/796
        dummy_line_status = 'LINEEND'
        yield token_features.get_layout_model_data([
            token_features.token_text,
            token_features.get_lower_token_text(),
            token_features.get_prefix(1),
            token_features.get_prefix(2),
            token_features.get_prefix(3),
            token_features.get_prefix(4),
            token_features.get_suffix(1),
            token_features.get_suffix(2),
            token_features.get_suffix(3),
            token_features.get_suffix(4),
            dummy_line_status,
            token_features.get_capitalisation_status_using_allcaps(),
            token_features.get_digit_status_using_containdigit(),
            token_features.get_str_is_single_char(),
            token_features.get_dummy_str_is_proper_name(),
            token_features.get_dummy_str_is_common_name(),
            token_features.get_str_is_first_name(),
            token_features.get_dummy_str_is_location_name(),
            token_features.get_str_is_country(),
            token_features.get_punctuation_type_feature(),
            token_features.get_word_shape_feature(),
            token_features.get_dummy_label()
        ])



import logging
from typing import Dict, Iterable, List, Optional, Union

from lxml import etree
from lxml.builder import ElementMaker

from sciencebeam_parser.utils.xml import get_text_content
from sciencebeam_parser.utils.xml_writer import parse_tag_expression
from sciencebeam_parser.document.layout_document import (
    LayoutBlock,
    LayoutPageCoordinates,
    LayoutToken
)
from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticMixedContentWrapper
)


LOGGER = logging.getLogger(__name__)


XML_NS = 'http://www.w3.org/XML/1998/namespace'
XML_NS_PREFIX = '{%s}' % XML_NS
XML_ID = XML_NS_PREFIX + 'id'

TEI_NS = 'http://www.tei-c.org/ns/1.0'
TEI_NS_PREFIX = '{%s}' % TEI_NS

TEI_NS_MAP = {
    'tei': TEI_NS
}

TEI_E = ElementMaker(namespace=TEI_NS, nsmap={
    None: TEI_NS
})


def get_or_create_element_at(parent: etree.ElementBase, path: List[str]) -> etree.ElementBase:
    if not path:
        return parent
    child = parent.find(TEI_NS_PREFIX + path[0])
    if child is None:
        LOGGER.debug('creating element: %s', path[0])
        tag_expression = parse_tag_expression(path[0])
        child = tag_expression.create_node(
            element_maker=TEI_E
        )
        parent.append(child)
    return get_or_create_element_at(child, path[1:])


def tei_xpath(parent: etree.ElementBase, xpath: str) -> List[etree.ElementBase]:
    return parent.xpath(xpath, namespaces=TEI_NS_MAP)


def get_tei_xpath_text_content_list(parent: etree.ElementBase, xpath: str) -> List[str]:
    return [get_text_content(node) for node in tei_xpath(parent, xpath)]


def get_required_styles(layout_token: LayoutToken) -> List[str]:
    required_styles = []
    if layout_token.font.is_bold:
        required_styles.append('bold')
    if layout_token.font.is_italics:
        required_styles.append('italic')
    if layout_token.font.is_subscript:
        required_styles.append('subscript')
    if layout_token.font.is_superscript:
        required_styles.append('superscript')
    return required_styles


def get_element_for_styles(styles: List[str], text: str) -> etree.ElementBase:
    if not styles:
        return text
    child: Optional[etree.ElementBase] = None
    for style in reversed(styles):
        LOGGER.debug('style: %r, child: %r, text: %r', style, child, text)
        if child is not None:
            child = TEI_E('hi', {'rend': style}, child)
        else:
            child = TEI_E('hi', {'rend': style}, text)
    return child


def format_coordinates(coordinates: LayoutPageCoordinates) -> str:
    return '%d,%.2f,%.2f,%.2f,%.2f' % (
        coordinates.page_number,
        coordinates.x,
        coordinates.y,
        coordinates.width,
        coordinates.height
    )


def format_coordinates_list(coordinates_list: List[LayoutPageCoordinates]) -> str:
    return ';'.join((
        format_coordinates(coordinates)
        for coordinates in coordinates_list
    ))


def get_default_attributes_for_layout_block(
    layout_block: LayoutBlock,
    enable_coordinates: bool = True
) -> Dict[str, str]:
    if enable_coordinates:
        formatted_coords = format_coordinates_list(
            layout_block.get_merged_coordinates_list()
        )
        if formatted_coords:
            return {'coords': formatted_coords}
    return {}


def iter_layout_block_tei_children(
    layout_block: LayoutBlock,
    enable_coordinates: bool = True
) -> Iterable[Union[str, etree.ElementBase]]:
    pending_styles: List[str] = []
    pending_text = ''
    pending_whitespace = ''
    if enable_coordinates:
        yield get_default_attributes_for_layout_block(
            layout_block=layout_block,
            enable_coordinates=enable_coordinates
        )
    for line in layout_block.lines:
        for token in line.tokens:
            required_styles = get_required_styles(token)
            LOGGER.debug('token: %r, required_styles=%r', token, required_styles)
            if required_styles != pending_styles:
                if pending_text:
                    yield get_element_for_styles(
                        pending_styles,
                        pending_text
                    )
                    pending_text = ''
                if pending_whitespace:
                    yield pending_whitespace
                    pending_whitespace = ''
                pending_styles = required_styles
            if pending_whitespace:
                pending_text += pending_whitespace
                pending_whitespace = ''
            pending_text += token.text
            pending_whitespace = token.whitespace
    if pending_text:
        yield get_element_for_styles(
            pending_styles,
            pending_text
        )


def extend_element(
    element: etree.ElementBase,
    children_or_attributes: Iterable[etree.ElementBase]
):
    for item in children_or_attributes:
        if isinstance(item, dict):
            element.attrib.update(item)
            continue
        if isinstance(item, str):
            try:
                previous_element = element[-1]
            except IndexError:
                previous_element = None
            if previous_element is not None:
                previous_element.tail = (
                    (previous_element.tail or '')
                    + item
                )
            else:
                element.text = (
                    (element.text or '')
                    + item
                )
            continue
        element.append(item)


def create_tei_note_element(
    note_type: str,
    layout_block: LayoutBlock
) -> etree.EntityBase:
    return TEI_E(
        'note',
        {'type': note_type},
        *iter_layout_block_tei_children(layout_block)
    )


def get_default_attributes_for_semantic_content(
    semantic_content: SemanticContentWrapper,
    **kwargs
) -> Dict[str, str]:
    attrib = get_default_attributes_for_layout_block(
        semantic_content.merged_block,
        **kwargs
    )
    if isinstance(semantic_content, SemanticMixedContentWrapper):
        if semantic_content.content_id:
            attrib = {
                **attrib,
                XML_ID: semantic_content.content_id
            }
    return attrib


def _create_tei_note_element(
    note_type: str,
    layout_block: LayoutBlock
) -> etree.EntityBase:
    return TEI_E(
        'note',
        {'type': note_type},
        *iter_layout_block_tei_children(layout_block)
    )


class TeiElementWrapper:
    def __init__(self, element: etree.ElementBase):
        self.element = element

    def xpath_nodes(self, xpath: str) -> List[etree.ElementBase]:
        return tei_xpath(self.element, xpath)

    def xpath(self, xpath: str) -> List['TeiElementWrapper']:
        return [TeiElementWrapper(node) for node in self.xpath_nodes(xpath)]

    def get_xpath_text_content_list(self, xpath: str) -> List[str]:
        return get_tei_xpath_text_content_list(self.element, xpath)

    def get_notes_text_list(self, note_type: str) -> List[str]:
        return get_tei_xpath_text_content_list(
            self.element,
            '//tei:note[@type="%s"]' % note_type,
        )

    def add_note(self, note_type: str, layout_block: LayoutBlock):
        self.element.append(create_tei_note_element(note_type, layout_block))


class TeiElementBuilder:
    def __init__(
        self,
        element: etree.ElementBase,
    ):
        self.element = element
        self.builder_by_path_fragment: Dict[str, 'TeiElementBuilder'] = {}

    def get_or_create(
        self,
        path: Optional[List[str]]
    ) -> 'TeiElementBuilder':
        if not path:
            return self
        key = path[0]
        builder = self.builder_by_path_fragment.get(key)
        if not builder:
            builder = TeiElementBuilder(TEI_E(key))
            self.element.append(builder.element)
            self.builder_by_path_fragment[key] = builder
        return builder.get_or_create(path[1:])

    def add_dict(self, attrib: dict):
        _attrib = self.element.attrib
        for k, v in attrib.items():
            _attrib[k] = v

    def append(
        self,
        child: Union[dict, etree.ElementBase]
    ):
        if isinstance(child, dict):
            self.add_dict(child)
            return
        self.element.append(child)

    def extend(self, children: List[Union[dict, etree.ElementBase]]):
        for child in children:
            self.append(child)



from abc import ABC, abstractmethod
import logging
from typing import (
    Dict,
    Iterable,
    List,
    Optional,
    Union
)

from lxml import etree

from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper
)


LOGGER = logging.getLogger(__name__)


T_ElementChildrenListItem = Union[dict, str, etree.ElementBase]
T_ElementChildrenList = List[T_ElementChildrenListItem]


class TeiElementFactoryContext(ABC):
    @abstractmethod
    def get_default_attributes_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        **kwargs
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def iter_layout_block_tei_children(
        self,
        layout_block: LayoutBlock,
        enable_coordinates: bool = True
    ) -> Iterable[Union[str, etree.ElementBase]]:
        pass

    @abstractmethod
    def get_tei_child_elements_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper
    ) -> List[etree.ElementBase]:
        pass

    @abstractmethod
    def append_tei_children_list_and_get_whitespace(
        self,
        children: T_ElementChildrenList,
        semantic_content: SemanticContentWrapper,
        pending_whitespace: str
    ) -> str:
        pass

    @abstractmethod
    def get_parent_path_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper
    ) -> Optional[List[str]]:
        pass


class TeiElementFactory(ABC):
    @abstractmethod
    def get_tei_children_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> T_ElementChildrenList:
        pass


class SingleElementTeiElementFactory(TeiElementFactory):
    @abstractmethod
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        pass

    def get_tei_children_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> T_ElementChildrenList:
        return [self.get_tei_element_for_semantic_content(
            semantic_content, context=context
        )]



import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union
)

from sciencebeam_trainer_delft.sequence_labelling.reader import load_data_crf_lines

from sciencebeam_parser.utils.labels import get_split_prefix_label, strip_tag_prefix

from sciencebeam_parser.document.layout_document import (
    LayoutToken,
    LayoutLine,
    LayoutBlock,
    LayoutPage,
    LayoutDocument
)
from sciencebeam_parser.models.data import (
    NEW_DOCUMENT_MARKER,
    AppFeaturesContext,
    DocumentFeaturesContext,
    LabeledLayoutModelData,
    LabeledLayoutToken,
    LayoutModelData,
    ModelDataGenerator,
    NewDocumentMarker
)
from sciencebeam_parser.models.extract import ModelSemanticExtractor
from sciencebeam_parser.models.training_data import TeiTrainingDataGenerator, TrainingTeiParser
from sciencebeam_parser.document.semantic_document import SemanticContentWrapper
from sciencebeam_parser.models.model_impl import ModelImpl, T_ModelImplFactory
from sciencebeam_parser.utils.lazy import LazyLoaded, Preloadable


LOGGER = logging.getLogger(__name__)


T = TypeVar('T')
U = TypeVar('U')


@dataclass
class LayoutModelLabel:
    label: str
    label_token_text: str
    layout_line: Optional[LayoutLine] = field(repr=False, default=None)
    layout_token: Optional[LayoutToken] = field(repr=False, default=None)


def iter_entities_including_other(seq: List[str]) -> Iterable[Tuple[str, int, int]]:
    """
    Similar to get_entities, but also other (`O`) tag
    """
    prev_tag = 'O'
    prev_start = 0
    for index, prefixed_tag in enumerate(seq):
        prefix, tag = get_split_prefix_label(prefixed_tag)
        if prefix == 'B' or tag != prev_tag:
            if prev_start < index:
                yield prev_tag, prev_start, index - 1
            prev_tag = tag
            prev_start = index
    if prev_start < len(seq):
        yield prev_tag, prev_start, len(seq) - 1


def get_entities_including_other(seq: List[str]) -> List[Tuple[str, int, int]]:
    return list(iter_entities_including_other(seq))


class LayoutDocumentLabelResult:
    def __init__(
        self,
        layout_document: LayoutDocument,
        layout_model_label_iterable: Iterable[LayoutModelLabel]
    ):
        self.layout_document = layout_document
        self.layout_model_label_list = list(layout_model_label_iterable)
        self.layout_document_labels_by_label: Dict[str, List[LayoutModelLabel]] = (
            defaultdict(list)
        )
        for layout_model_label in self.layout_model_label_list:
            tag_without_prefix = strip_tag_prefix(layout_model_label.label)
            self.layout_document_labels_by_label[tag_without_prefix].append(
                layout_model_label
            )

    def get_available_labels(self) -> Set[str]:
        return set(self.layout_document_labels_by_label.keys())

    def get_layout_document_labels_by_labels(self, labels: List[str]) -> List[LayoutModelLabel]:
        if not labels:
            return []
        if len(labels) == 1:
            return self.layout_document_labels_by_label.get(labels[0], [])
        result: List[LayoutModelLabel] = []
        for label in labels:
            result.extend(self.layout_document_labels_by_label.get(label, []))
        return result

    def get_filtered_document_by_label(self, label: str) -> LayoutDocument:
        return self.get_filtered_document_by_labels([label])

    def get_filtered_document_by_labels(
        self,
        labels: List[str]
    ):  # pylint: disable=too-many-branches
        layout_document = LayoutDocument(pages=[])
        layout_document_labels = self.get_layout_document_labels_by_labels(labels)
        if not layout_document_labels:
            LOGGER.warning(
                'no layout_lines_to_include found for: %r, available keys=%r',
                labels, self.layout_document_labels_by_label.keys()
            )
            return layout_document
        layout_token_ids_to_include = {
            id(layout_document_label.layout_token)
            for layout_document_label in layout_document_labels
            if layout_document_label.layout_token
        }
        LOGGER.debug('layout_tokens_to_include: %s', layout_token_ids_to_include)
        layout_line_ids_to_include: Set[int] = set()
        if not layout_token_ids_to_include:
            layout_line_ids_to_include = {
                id(layout_document_label.layout_line)
                for layout_document_label in layout_document_labels
                if layout_document_label.layout_line
            }
        LOGGER.debug('layout_line_ids_to_include: %s', layout_line_ids_to_include)
        result_page: Optional[LayoutPage] = None
        for page in self.layout_document.pages:  # pylint: disable=too-many-nested-blocks
            result_page = None
            result_block: Optional[LayoutBlock] = None
            for block in page.blocks:
                result_block = None
                for line in block.lines:
                    accepted_line: Optional[LayoutLine] = None
                    if layout_token_ids_to_include:
                        accepted_tokens: List[LayoutToken] = []
                        for token in line.tokens:
                            if id(token) in layout_token_ids_to_include:
                                accepted_tokens.append(token)
                        if not accepted_tokens:
                            continue
                        if len(line.tokens) == accepted_tokens:
                            accepted_line = line
                        else:
                            accepted_line = LayoutLine(tokens=accepted_tokens)
                    else:
                        if id(line) not in layout_line_ids_to_include:
                            continue
                        accepted_line = line
                    if result_page is None:
                        result_page = LayoutPage(blocks=[])
                        layout_document.pages.append(result_page)
                    if result_block is None:
                        result_block = LayoutBlock(lines=[])
                        result_page.blocks.append(result_block)
                    result_block.lines.append(accepted_line)
        return layout_document


def iter_entity_layout_blocks_for_labeled_layout_tokens(
    labeled_layout_tokens: Iterable[LabeledLayoutToken]
) -> Iterable[Tuple[str, LayoutBlock]]:
    layout_tokens = [result.layout_token for result in labeled_layout_tokens]
    labels = [result.label for result in labeled_layout_tokens]
    LOGGER.debug('layout_tokens: %s', layout_tokens)
    LOGGER.debug('labels: %s', labels)
    for tag, start, end in get_entities_including_other(list(labels)):
        yield tag, LayoutBlock.for_tokens(layout_tokens[start:end + 1])


def iter_entity_values_predicted_labels(
    tag_result: List[Tuple[str, str]]
) -> Iterable[Tuple[str, str]]:
    tokens, labels = zip(*tag_result)
    LOGGER.debug('tokens: %s', tokens)
    LOGGER.debug('labels: %s', labels)
    for tag, start, end in get_entities_including_other(list(labels)):
        yield tag, ' '.join(tokens[start:end + 1])


def iter_labeled_layout_token_for_layout_model_label(
    layout_model_label_iterable: Iterable[LayoutModelLabel]
) -> Iterable[LabeledLayoutToken]:
    for layout_model_label in layout_model_label_iterable:
        layout_token = layout_model_label.layout_token
        assert layout_token is not None
        yield LabeledLayoutToken(
            layout_model_label.label,
            layout_token
        )


def iter_data_lines_for_model_data_iterables(
    model_data_iterables: Iterable[Iterable[LayoutModelData]]
) -> Iterable[str]:
    for index, model_data_list in enumerate(model_data_iterables):
        if index > 0:
            yield ''
        for model_data in model_data_list:
            yield model_data.data_line


class Model(ABC, Preloadable):
    def __init__(
        self,
        model_impl_factory: Optional[T_ModelImplFactory],
        model_config: Optional[dict] = None
    ) -> None:
        self._model_impl_factory = model_impl_factory
        self._lazy_model_impl = LazyLoaded[ModelImpl](self._load_model_impl)
        self.model_config = model_config or {}

    def __repr__(self) -> str:
        return '%s(model_config=%r, loaded=%r)' % (
            type(self).__name__, self.model_config, self._lazy_model_impl.is_loaded
        )

    @abstractmethod
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> ModelDataGenerator:
        pass

    # @abstractmethod
    def get_semantic_extractor(self) -> ModelSemanticExtractor:
        raise NotImplementedError()

    # @abstractmethod
    def get_tei_training_data_generator(self) -> TeiTrainingDataGenerator:
        raise NotImplementedError()

    # @abstractmethod
    def get_training_tei_parser(self) -> TrainingTeiParser:
        raise NotImplementedError()

    def _load_model_impl(self) -> ModelImpl:
        assert self._model_impl_factory, 'model impl factory required'
        LOGGER.info('creating model impl: %r', self._model_impl_factory)
        model_impl = self._model_impl_factory()
        if not isinstance(model_impl, ModelImpl):
            raise TypeError('invalid model impl type: %r' % model_impl)
        return model_impl

    @property
    def model_impl(self) -> ModelImpl:
        was_loaded = self._lazy_model_impl.is_loaded
        model_impl = self._lazy_model_impl.get()
        if was_loaded:
            LOGGER.info('model impl already loaded: %r', model_impl)
        return model_impl

    def preload(self):
        model_impl = self.model_impl
        model_impl.preload()

    def iter_semantic_content_for_entity_blocks(
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        return self.get_semantic_extractor().iter_semantic_content_for_entity_blocks(
            entity_tokens,
            **kwargs
        )

    def predict_labels(
        self,
        texts: List[List[str]],
        features: List[List[List[str]]],
        output_format: Optional[str] = None
    ) -> List[List[Tuple[str, str]]]:
        return self.model_impl.predict_labels(texts, features, output_format)

    def _iter_flat_label_model_data_lists_to(  # pylint: disable=too-many-locals
        self,
        model_data_list_iterable: Iterable[Sequence[LayoutModelData]],
        item_factory: Callable[[str, LayoutModelData], T]
    ) -> Iterable[Union[T, NewDocumentMarker]]:
        # Note: currently we do need a list
        model_data_lists = list(model_data_list_iterable)
        if not model_data_lists:
            return
        data_lines = list(iter_data_lines_for_model_data_iterables(
            model_data_lists
        ))
        texts, features = load_data_crf_lines(data_lines)
        texts = texts.tolist()
        tag_result = self.predict_labels(
            texts=texts, features=features, output_format=None
        )
        if not tag_result:
            return
        if len(tag_result) != len(model_data_lists):
            raise AssertionError('tag result does not match number of docs: %d != %d' % (
                len(tag_result), len(model_data_lists)
            ))
        for index, (doc_tag_result, model_data_list) in enumerate(
            zip(tag_result, model_data_lists)
        ):
            if index > 0:
                yield NEW_DOCUMENT_MARKER
            if len(doc_tag_result) != len(model_data_list):
                raise AssertionError('doc tag result does not match data: %d != %d' % (
                    len(doc_tag_result), len(model_data_list)
                ))
            for token_tag_result, token_model_data in zip(doc_tag_result, model_data_list):
                label_token_text, token_label = token_tag_result
                if label_token_text != token_model_data.label_token_text:
                    raise AssertionError(
                        f'actual: {repr(label_token_text)}'
                        f', expected: {repr(token_model_data.label_token_text)}'
                    )
                yield item_factory(
                    token_label,
                    token_model_data
                )

    def _iter_stacked_label_model_data_lists_to(
        self,
        model_data_list_iterable: Iterable[Sequence[LayoutModelData]],
        item_factory: Callable[[str, LayoutModelData], T]
    ) -> Iterable[Sequence[T]]:
        # Note: currently we do need a list
        model_data_lists = list(model_data_list_iterable)
        if not model_data_lists:
            return
        doc_items: List[T] = []
        result_doc_count = 0
        for item in self._iter_flat_label_model_data_lists_to(
            model_data_lists,
            item_factory=item_factory
        ):
            if isinstance(item, NewDocumentMarker):
                yield doc_items
                doc_items = []
                result_doc_count += 1
                continue
            doc_items.append(item)
        if result_doc_count < len(model_data_lists):
            yield doc_items

    def iter_label_layout_documents(
        self,
        layout_documents: List[LayoutDocument],
        app_features_context: AppFeaturesContext
    ) -> Iterable[List[LayoutModelLabel]]:
        doc_layout_model_labels: List[LayoutModelLabel] = []
        result_doc_count = 0
        for layout_model_label in self._iter_label_layout_documents(
            layout_documents,
            app_features_context=app_features_context
        ):
            if isinstance(layout_model_label, NewDocumentMarker):
                yield doc_layout_model_labels
                doc_layout_model_labels = []
                result_doc_count += 1
                continue
            doc_layout_model_labels.append(layout_model_label)
        if result_doc_count < len(layout_documents):
            yield doc_layout_model_labels

    def iter_label_layout_document(
        self,
        layout_document: LayoutDocument,
        app_features_context: AppFeaturesContext
    ) -> Iterable[LayoutModelLabel]:
        for layout_model_label in self._iter_label_layout_documents(
            [layout_document],
            app_features_context=app_features_context
        ):
            assert isinstance(layout_model_label, LayoutModelLabel)
            yield layout_model_label

    def _iter_label_layout_documents(  # pylint: disable=too-many-locals
        self,
        layout_documents: Iterable[LayoutDocument],
        app_features_context: AppFeaturesContext
    ) -> Iterable[Union[LayoutModelLabel, NewDocumentMarker]]:
        data_generator = self.get_data_generator(
            document_features_context=DocumentFeaturesContext(
                app_features_context=app_features_context
            )
        )
        model_data_lists = [
            list(data_generator.iter_model_data_for_layout_document(
                layout_document
            ))
            for layout_document in layout_documents
        ]
        return self._iter_flat_label_model_data_lists_to(
            model_data_lists,
            lambda label, model_data: LayoutModelLabel(
                label=label,
                label_token_text=model_data.label_token_text,
                layout_line=model_data.layout_line,
                layout_token=model_data.layout_token
            )
        )

    def iter_labeled_model_data_list_for_model_data_list_iterable(
        self,
        model_data_list_iterable: Iterable[Sequence[LayoutModelData]]
    ) -> Iterable[Sequence[LabeledLayoutModelData]]:
        return self._iter_stacked_label_model_data_lists_to(
            model_data_list_iterable,
            lambda label, model_data: LabeledLayoutModelData.from_model_data(
                model_data,
                label=label
            )
        )

    def get_label_layout_document_result(
        self,
        layout_document: LayoutDocument,
        app_features_context: AppFeaturesContext
    ) -> LayoutDocumentLabelResult:
        return LayoutDocumentLabelResult(
            layout_document=layout_document,
            layout_model_label_iterable=self.iter_label_layout_document(
                layout_document,
                app_features_context=app_features_context
            )
        )

    def iter_predict_labels_for_layout_document(
        self,
        layout_document: LayoutDocument,
        app_features_context: AppFeaturesContext
    ) -> Iterable[LabeledLayoutToken]:
        # Note: this should get merged with Model.iter_label_layout_document
        yield from iter_labeled_layout_token_for_layout_model_label(
            self.iter_label_layout_document(
                layout_document,
                app_features_context=app_features_context
            )
        )

    def predict_labels_for_layout_document(
        self,
        layout_document: LayoutDocument,
        app_features_context: AppFeaturesContext
    ) -> List[LabeledLayoutToken]:
        return list(self.iter_predict_labels_for_layout_document(
            layout_document,
            app_features_context=app_features_context
        ))

    def predict_labels_for_layout_documents(
        self,
        layout_documents: List[LayoutDocument],
        app_features_context: AppFeaturesContext
    ) -> List[List[LabeledLayoutToken]]:
        return [
            list(iter_labeled_layout_token_for_layout_model_label(
                layout_model_labels
            ))
            for layout_model_labels in self.iter_label_layout_documents(
                layout_documents,
                app_features_context=app_features_context
            )
        ]

    def iter_entity_layout_blocks_for_labeled_layout_tokens(
        self,
        labeled_layout_tokens: Iterable[LabeledLayoutToken]
    ) -> Iterable[Tuple[str, LayoutBlock]]:
        return iter_entity_layout_blocks_for_labeled_layout_tokens(labeled_layout_tokens)

    def iter_semantic_content_for_labeled_layout_tokens(
        self,
        labeled_layout_tokens: Iterable[LabeledLayoutToken],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        return self.iter_semantic_content_for_entity_blocks(
            self.iter_entity_layout_blocks_for_labeled_layout_tokens(
                labeled_layout_tokens
            ),
            **kwargs
        )



import PIL.Image


def get_image_with_max_resolution(
    image: PIL.Image.Image,
    max_resolution: int
) -> PIL.Image.Image:
    if max(image.width, image.height) <= max_resolution:
        return image
    if image.width > image.height:
        target_width = max_resolution
        target_height = max(1, round(image.height / image.width * target_width))
    else:
        target_height = max_resolution
        target_width = max(1, round(image.width / image.height * target_height))
    return image.resize((target_width, target_height))



import logging

from lxml.builder import ElementMaker

from sciencebeam_parser.models.training_data import (
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


TEI_E = ElementMaker()


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/HeaderParser.java
ROOT_TRAINING_XML_ELEMENT_PATH = ['text', 'front']

TRAINING_XML_ELEMENT_PATH_BY_LABEL_WITHOUT_ALIAS = {
    '<title>': ROOT_TRAINING_XML_ELEMENT_PATH + ['docTitle', 'titlePart'],
    '<author>': ROOT_TRAINING_XML_ELEMENT_PATH + ['byline', 'docAuthor'],
    '<address>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address'],
    '<date>': ROOT_TRAINING_XML_ELEMENT_PATH + ['date'],
    '<page>': ROOT_TRAINING_XML_ELEMENT_PATH + ['page'],
    '<publisher>': ROOT_TRAINING_XML_ELEMENT_PATH + ['publisher'],
    '<journal>': ROOT_TRAINING_XML_ELEMENT_PATH + ['journal'],
    '<affiliation>': ROOT_TRAINING_XML_ELEMENT_PATH + ['byline', 'affiliation'],
    '<note>': ROOT_TRAINING_XML_ELEMENT_PATH,
    '<abstract>': ROOT_TRAINING_XML_ELEMENT_PATH + ['div[@type="abstract"]'],
    '<email>': ROOT_TRAINING_XML_ELEMENT_PATH + ['email'],
    '<pubnum>': ROOT_TRAINING_XML_ELEMENT_PATH + ['idno'],
    '<keyword>': ROOT_TRAINING_XML_ELEMENT_PATH + ['keyword'],
    '<phone>': ROOT_TRAINING_XML_ELEMENT_PATH + ['phone'],
    '<web>': ROOT_TRAINING_XML_ELEMENT_PATH + ['ptr[@type="web"]'],
    '<meeting>': ROOT_TRAINING_XML_ELEMENT_PATH + ['meeting'],
    '<submission>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="submission"]'],
    '<reference>': ROOT_TRAINING_XML_ELEMENT_PATH + ['reference'],
    '<copyright>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="copyright"]'],
    '<funding>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="funding"]'],
    '<doctype>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="doctype"]'],
    '<group>': ROOT_TRAINING_XML_ELEMENT_PATH + ['note[@type="group"]']
}


TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    **TRAINING_XML_ELEMENT_PATH_BY_LABEL_WITHOUT_ALIAS,
    '<location>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address'],
    '<institution>': ROOT_TRAINING_XML_ELEMENT_PATH + ['byline', 'affiliation']
}


class HeaderTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.header.tei.xml'
    DEFAULT_DATA_FILENAME_SUFFIX = '.header'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            element_maker=TEI_E,
            default_tei_filename_suffix=(
                HeaderTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=(
                HeaderTeiTrainingDataGenerator.DEFAULT_DATA_FILENAME_SUFFIX
            ),
            default_tei_sub_directory='header/corpus/tei',
            default_data_sub_directory='header/corpus/raw'
        )


class HeaderTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL_WITHOUT_ALIAS
            ),
            use_tei_namespace=False
        )



from dataclasses import dataclass
from typing import Optional

from sciencebeam_trainer_delft.utils.download_manager import DownloadManager


DEFAULT_WAPITI_PATH = 'wapiti'


def install_wapiti_and_get_path_or_none(
    install_url: Optional[str],
    download_manager: DownloadManager
) -> Optional[str]:
    from sciencebeam_trainer_delft.sequence_labelling.engines.wapiti_install import (  # noqa pylint: disable=import-outside-toplevel
        install_wapiti_and_get_path_or_none as _install_wapiti_and_get_path_or_none
    )
    return _install_wapiti_and_get_path_or_none(
        install_url,
        download_manager=download_manager
    )


@dataclass
class LazyWapitiBinaryWrapper:
    download_manager: DownloadManager
    install_url: Optional[str] = None
    _binary_path: Optional[str] = None

    def get_binary_path(self) -> str:
        if not self.install_url:
            return DEFAULT_WAPITI_PATH
        if self._binary_path:
            return self._binary_path
        self._binary_path = install_wapiti_and_get_path_or_none(
            self.install_url,
            download_manager=self.download_manager
        ) or DEFAULT_WAPITI_PATH
        return self._binary_path



import functools
import logging
import os
from abc import ABC, abstractmethod
from typing import Counter, Iterable, List, Optional, Sequence

from sciencebeam_parser.utils.bounding_box import BoundingBox
from sciencebeam_parser.document.layout_document import (
    LayoutBlock,
    LayoutDocument,
    LayoutGraphic,
    LayoutPage,
    LayoutPageCoordinates,
    LayoutToken
)
from sciencebeam_parser.document.semantic_document import SemanticContentWrapper, SemanticGraphic


LOGGER = logging.getLogger(__name__)


class DocumentGraphicProvider(ABC):
    @abstractmethod
    def iter_semantic_graphic_for_layout_document(
        self,
        layout_document: LayoutDocument,
        extract_graphic_assets: bool
    ) -> Iterable[SemanticGraphic]:
        pass


def get_semantic_graphic_for_layout_graphic(
    layout_graphic: LayoutGraphic,
    extract_graphic_assets: bool
) -> SemanticGraphic:
    relative_path: Optional[str] = None
    if layout_graphic.local_file_path and extract_graphic_assets:
        relative_path = os.path.basename(layout_graphic.local_file_path)
    return SemanticGraphic(
        layout_graphic=layout_graphic,
        relative_path=relative_path
    )


def get_semantic_graphic_list_for_layout_graphic_list(
    layout_graphic_iterable: Iterable[LayoutGraphic],
    extract_graphic_assets: bool
) -> List[SemanticGraphic]:
    return [
        get_semantic_graphic_for_layout_graphic(
            layout_graphic,
            extract_graphic_assets=extract_graphic_assets
        )
        for layout_graphic in layout_graphic_iterable
        if layout_graphic.coordinates
    ]


def get_page_numbers_for_semantic_content_list(
    semantic_content_list: Sequence[SemanticContentWrapper]
) -> Sequence[int]:
    return sorted({
        coordinates.page_number
        for semantic_content in semantic_content_list
        for coordinates in semantic_content.merged_block.get_merged_coordinates_list()
    })


def get_all_page_numbers_of_layout_document(
    layout_document: LayoutDocument
) -> Sequence[int]:
    return sorted({
        page.meta.page_number
        for page in layout_document.pages
    })


def get_graphic_matching_candidate_page_numbers_for_semantic_content_list(
    semantic_content_list: Sequence[SemanticContentWrapper],
    layout_document: Optional[LayoutDocument] = None
) -> Sequence[int]:
    page_numbers = get_page_numbers_for_semantic_content_list(semantic_content_list)
    if layout_document:
        document_page_numbers = set(get_all_page_numbers_of_layout_document(layout_document))
        page_numbers = sorted(
            set(page_numbers).union({
                page_number + 1
                for page_number in page_numbers
                if page_number + 1 in document_page_numbers
            })
        )
    return page_numbers


def get_page_numbers_with_uncommon_page_dimension(
    layout_document: LayoutDocument
) -> Sequence[int]:
    page_dimension_counter = Counter((
        page.meta.coordinates.bounding_box
        for page in layout_document.pages
        if page.meta and page.meta.coordinates
    ))
    LOGGER.debug('page_dimension_counter: %r', page_dimension_counter)
    if len(page_dimension_counter) < 2:
        return []
    most_common_page_dimension = page_dimension_counter.most_common(1)[0][0]
    LOGGER.debug('most_common_page_dimension: %r', most_common_page_dimension)
    return sorted({
        page.meta.page_number
        for page in layout_document.pages
        if (
            page.meta
            and page.meta.coordinates
            and page.meta.coordinates.bounding_box != most_common_page_dimension
        )
    })


def is_page_with_mostly_bitmap_graphics(
    layout_page: LayoutPage
) -> bool:
    if not layout_page.meta or not layout_page.meta.coordinates:
        LOGGER.debug('page has no coordinates')
        return False
    page_area = layout_page.meta.coordinates.bounding_box.area
    if not page_area:
        LOGGER.debug('page has no area')
        return False
    bitmap_graphics_with_area_ratio = [
        (graphic, graphic.coordinates.bounding_box.area / page_area)
        for graphic in layout_page.graphics
        if (
            graphic.graphic_type != 'svg'
            and graphic.coordinates
        )
    ]
    LOGGER.debug('bitmap_graphics_with_area_ratio: %r', bitmap_graphics_with_area_ratio)
    if not bitmap_graphics_with_area_ratio:
        LOGGER.debug('no bitmap images')
        return False
    accepted_bitmap_graphics = [
        bitmap_graphics
        for bitmap_graphics, area_ratio in bitmap_graphics_with_area_ratio
        if area_ratio > 0.5
    ]
    if not accepted_bitmap_graphics:
        LOGGER.debug('no too small bitmap images')
        return False
    return True


def get_page_numbers_with_mostly_bitmap_graphics(
    layout_document: LayoutDocument
) -> Sequence[int]:
    return [
        page.meta.page_number
        for page in layout_document.pages
        if (
            page.meta
            and is_page_with_mostly_bitmap_graphics(page)
        )
    ]


def are_page_coordinates_within_bounding_box(
    page_coordinates: Optional[LayoutPageCoordinates],
    bounding_box: BoundingBox,
    min_area_ratio: float = 0.5
) -> bool:
    if not page_coordinates:
        return False
    item_bounding_box = page_coordinates.bounding_box
    item_bounding_box_area = item_bounding_box.area
    if not item_bounding_box_area:
        return False
    intersection_bounding_box = item_bounding_box.intersection(bounding_box)
    if not intersection_bounding_box:
        return False
    if intersection_bounding_box.area / item_bounding_box_area < min_area_ratio:
        return False
    return True


def is_layout_token_within_bounding_box(layout_token: LayoutToken, **kwargs) -> bool:
    return are_page_coordinates_within_bounding_box(layout_token.coordinates, **kwargs)


def is_layout_graphic_within_bounding_box(layout_graphic: LayoutGraphic, **kwargs) -> bool:
    return are_page_coordinates_within_bounding_box(layout_graphic.coordinates, **kwargs)


def _remove_tokens_within_bounding_box_flatmap_fn(
    layout_token: LayoutToken,
    **kwargs
) -> List[LayoutToken]:
    if not is_layout_token_within_bounding_box(layout_token, **kwargs):
        return [layout_token]
    return []


def get_layout_page_with_text_or_graphic_replaced_by_graphic(
    layout_page: LayoutPage,
    semantic_graphic: SemanticGraphic,
    is_only_semantic_graphic_on_page: bool,
    is_replace_overlapping_text: bool
) -> LayoutPage:
    layout_graphic = semantic_graphic.layout_graphic
    assert layout_graphic
    assert layout_graphic.coordinates
    graphic_bounding_box = layout_graphic.coordinates.bounding_box
    if is_only_semantic_graphic_on_page:
        layout_graphic = layout_graphic._replace(
            related_block=LayoutBlock.for_tokens(list(layout_page.iter_all_tokens()))
        )
    modified_layout_page = (
        layout_page.replace(
            graphics=[
                _layout_graphic
                for _layout_graphic in layout_page.graphics
                if not is_layout_graphic_within_bounding_box(
                    _layout_graphic,
                    bounding_box=graphic_bounding_box
                )
             ] + [layout_graphic]
        )
    )
    if is_replace_overlapping_text:
        modified_layout_page = (
            modified_layout_page
            .flat_map_layout_tokens(functools.partial(
                _remove_tokens_within_bounding_box_flatmap_fn,
                bounding_box=graphic_bounding_box
            )).remove_empty_blocks()
        )
    return modified_layout_page


def get_layout_document_with_text_or_graphic_replaced_by_graphics(
    layout_document: LayoutDocument,
    semantic_graphics: Iterable[SemanticGraphic],
    is_replace_overlapping_text: bool
) -> LayoutDocument:
    page_by_page_number = {
        page.meta.page_number: page
        for page in layout_document.pages
        if page.meta
    }
    LOGGER.debug('page_by_page_number.keys: %r', page_by_page_number.keys())
    has_changes = False
    semantic_graphics_list = list(semantic_graphics)
    semantic_graphic_count_by_page = Counter((
        semantic_graphic.layout_graphic.coordinates.page_number
        for semantic_graphic in semantic_graphics_list
        if (
            semantic_graphic.layout_graphic
            and semantic_graphic.layout_graphic.coordinates
        )
    ))
    for semantic_graphic in semantic_graphics_list:
        layout_graphic = semantic_graphic.layout_graphic
        assert layout_graphic
        if not layout_graphic.coordinates:
            continue
        page_number = layout_graphic.coordinates.page_number
        page_by_page_number[page_number] = (
            get_layout_page_with_text_or_graphic_replaced_by_graphic(
                page_by_page_number[page_number],
                semantic_graphic,
                is_only_semantic_graphic_on_page=(
                    semantic_graphic_count_by_page[page_number] < 2
                ),
                is_replace_overlapping_text=is_replace_overlapping_text
            )
        )
        has_changes = True
    if not has_changes:
        return layout_document
    pages = [
        (
            page_by_page_number[page.meta.page_number]
            if page.meta
            else page
        )
        for page in layout_document.pages
    ]
    return layout_document.replace(pages=pages)


def get_layout_document_with_text_and_graphics_replaced_by_graphics(
    layout_document: LayoutDocument,
    semantic_graphics: Iterable[SemanticGraphic]
) -> LayoutDocument:
    return get_layout_document_with_text_or_graphic_replaced_by_graphics(
        layout_document,
        semantic_graphics=semantic_graphics,
        is_replace_overlapping_text=True
    )


def get_layout_document_with_graphics_replaced_by_graphics(
    layout_document: LayoutDocument,
    semantic_graphics: Iterable[SemanticGraphic]
) -> LayoutDocument:
    return get_layout_document_with_text_or_graphic_replaced_by_graphics(
        layout_document,
        semantic_graphics=semantic_graphics,
        is_replace_overlapping_text=False
    )


class SimpleDocumentGraphicProvider(DocumentGraphicProvider):
    def iter_semantic_graphic_for_layout_document(
        self,
        layout_document: LayoutDocument,
        extract_graphic_assets: bool
    ) -> Iterable[SemanticGraphic]:
        return get_semantic_graphic_list_for_layout_graphic_list(
            layout_document.iter_all_graphics(),
            extract_graphic_assets=extract_graphic_assets
        )



import logging
from typing import List, Sequence, Tuple

import PIL.Image

from layoutparser.elements.layout import Layout
from layoutparser.models.auto_layoutmodel import AutoLayoutModel
from layoutparser.models.base_layoutmodel import BaseLayoutModel

from sciencebeam_parser.utils.bounding_box import BoundingBox
from sciencebeam_parser.cv_models.cv_model import (
    ComputerVisionModel,
    ComputerVisionModelInstance,
    ComputerVisionModelResult,
    SimpleComputerVisionModelInstance
)
from sciencebeam_parser.utils.lazy import LazyLoaded


LOGGER = logging.getLogger(__name__)

DEFAULT_MODEL_PATH = 'lp://efficientdet/PubLayNet'

DEFAULT_SCORE_THRESHOLD = 0.0


def load_model(model_path: str) -> BaseLayoutModel:
    LOGGER.info('loading model: %r', model_path)
    return AutoLayoutModel(model_path)


def get_bounding_box_for_layout_parser_coordinates(
    coordinates: Tuple[float, float, float, float]
) -> BoundingBox:
    x1, y1, x2, y2 = coordinates
    return BoundingBox(x=x1, y=y1, width=x2 - x1, height=y2 - y1)


def is_bounding_box_overlapping_with_any_bounding_boxes(
    bounding_box: BoundingBox,
    other_bounding_boxes: Sequence[BoundingBox],
    max_overlap_ratio: float = 0.1
) -> bool:
    bounding_box_area = bounding_box.area
    for other_bounding_box in other_bounding_boxes:
        intersection_bounding_box = bounding_box.intersection(
            other_bounding_box
        )
        if not intersection_bounding_box:
            continue
        if intersection_bounding_box.area / bounding_box_area >= max_overlap_ratio:
            return True
    return False


class LayoutParserComputerVisionModelResult(ComputerVisionModelResult):
    def __init__(
        self,
        layout: Layout,
        score_threshold: float,
        avoid_overlapping: bool,
        max_overlap_ratio: float = 0.1
    ):
        super().__init__()
        self.layout = layout
        self.score_threshold = score_threshold
        self.avoid_overlapping = avoid_overlapping
        self.max_overlap_ratio = max_overlap_ratio
        LOGGER.debug('layout: %r', layout)

    def get_instances_by_type_names(
        self,
        type_names: Sequence[str]
    ) -> Sequence[ComputerVisionModelInstance]:
        instances = [
            SimpleComputerVisionModelInstance(
                bounding_box=get_bounding_box_for_layout_parser_coordinates(block.coordinates),
                type_name=block.type
            )
            for block in self.layout
            if (
                block.type in type_names
                and block.score >= self.score_threshold
            )
        ]
        instances = [
            instance
            for instance in instances
            if instance.get_bounding_box()
        ]
        if self.avoid_overlapping:
            _instances = instances
            instances = []
            prev_bounding_boxes: List[BoundingBox] = []
            for instance in _instances:
                bounding_box = instance.get_bounding_box()
                if is_bounding_box_overlapping_with_any_bounding_boxes(
                    bounding_box,
                    prev_bounding_boxes,
                    max_overlap_ratio=self.max_overlap_ratio
                ):
                    LOGGER.debug(
                        'bounding box overlapping with prev: %r ~ %r',
                        bounding_box, prev_bounding_boxes
                    )
                    continue
                instances.append(instance)
                prev_bounding_boxes.append(bounding_box)
        return instances


class LayoutParserComputerVisionModel(ComputerVisionModel):
    def __init__(
        self,
        config: dict,
        model_path: str = DEFAULT_MODEL_PATH,
    ):
        super().__init__()
        self.score_threshold = float(config.get('score_threshold', DEFAULT_SCORE_THRESHOLD))
        self.avoid_overlapping = bool(config.get('avoid_overlapping', True))
        self.model_path = model_path
        self._lazy_model = LazyLoaded[BaseLayoutModel](self._load_model)

    def _load_model(self) -> BaseLayoutModel:
        model = load_model(self.model_path)
        LOGGER.info('loaded layout model: %r', self.model_path)
        return model

    @property
    def layout_model(self) -> BaseLayoutModel:
        return self._lazy_model.get()

    def preload(self):
        self._lazy_model.get()

    def predict_single(self, image: PIL.Image.Image) -> ComputerVisionModelResult:
        return LayoutParserComputerVisionModelResult(
            self.layout_model.detect(image),
            score_threshold=self.score_threshold,
            avoid_overlapping=self.avoid_overlapping
        )



import logging
from typing import Iterable, Mapping, Optional, Tuple

from sciencebeam_parser.document.semantic_document import (
    SemanticCaption,
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticFigure,
    SemanticLabel
)
from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.models.extract import SimpleModelSemanticExtractor


LOGGER = logging.getLogger(__name__)


SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<label>': SemanticLabel,
    '<figDesc>': SemanticCaption
}


class FigureSemanticExtractor(SimpleModelSemanticExtractor):
    def __init__(self):
        super().__init__(semantic_content_class_by_tag=SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG)

    def iter_semantic_content_for_entity_blocks(  # pylint: disable=arguments-differ
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        figure: Optional[SemanticFigure] = None
        for name, layout_block in entity_tokens:
            if not figure:
                figure = SemanticFigure()
            semantic_content = self.get_semantic_content_for_entity_name(
                name, layout_block=layout_block
            )
            figure.add_content(semantic_content)
        if figure:
            yield figure



import logging

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticAuthor,
    SemanticContentWrapper,
    SemanticDate,
    SemanticLabel,
    SemanticRawReference,
    SemanticRawReferenceText,
    SemanticReference,
    SemanticReferenceList,
    SemanticTitle
)
from sciencebeam_parser.document.tei.common import (
    TEI_E,
    TeiElementBuilder,
    create_tei_note_element
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)
from sciencebeam_parser.document.tei.author import get_tei_author_for_semantic_author_element


LOGGER = logging.getLogger(__name__)


def _get_tei_raw_reference_element(
    semantic_raw_ref: SemanticRawReference,
    context: TeiElementFactoryContext
) -> etree.ElementBase:
    LOGGER.debug('semantic_raw_ref: %s', semantic_raw_ref)
    children = []
    for semantic_content in semantic_raw_ref:
        if isinstance(semantic_content, SemanticRawReferenceText):
            children.append(create_tei_note_element(
                'raw_reference', semantic_content.merged_block
            ))
            continue
        children.extend(context.get_tei_child_elements_for_semantic_content(semantic_content))
    tei_ref = TEI_E(
        'biblStruct',
        context.get_default_attributes_for_semantic_content(semantic_raw_ref),
        *children
    )
    return tei_ref


def get_tei_reference_element(  # pylint: disable=too-many-branches
    semantic_ref: SemanticReference,
    context: TeiElementFactoryContext
) -> etree.ElementBase:
    LOGGER.debug('semantic_ref: %s', semantic_ref)
    tei_ref = TeiElementBuilder(TEI_E(
        'biblStruct',
        context.get_default_attributes_for_semantic_content(semantic_ref)
    ))
    is_first_date = True
    for semantic_content in semantic_ref:
        parent_path = context.get_parent_path_for_semantic_content(
            semantic_content
        )
        tei_child_parent = tei_ref.get_or_create(parent_path)
        if isinstance(semantic_content, SemanticLabel):
            tei_child_parent.append(create_tei_note_element(
                'label', semantic_content.merged_block
            ))
            continue
        if isinstance(semantic_content, SemanticRawReferenceText):
            tei_child_parent.append(create_tei_note_element(
                'raw_reference', semantic_content.merged_block
            ))
            continue
        if isinstance(semantic_content, SemanticTitle):
            tei_child_parent.append(TEI_E(
                'title',
                {'level': 'a', 'type': 'main'},
                *context.iter_layout_block_tei_children(
                    semantic_content.merged_block
                )
            ))
            continue
        if isinstance(semantic_content, SemanticAuthor):
            tei_child_parent.append(get_tei_author_for_semantic_author_element(
                semantic_content,
                context=context
            ))
            continue
        if isinstance(semantic_content, SemanticDate):
            tei_child_parent = tei_ref.get_or_create(['monogr', 'imprint'])
            attrib = {}
            if is_first_date:
                # assume first date is published date (more or less matches GROBID)
                attrib['type'] = 'published'
            if semantic_content.year:
                attrib['when'] = str(semantic_content.year)
            tei_child_parent.append(TEI_E(
                'date', attrib,
                *context.iter_layout_block_tei_children(layout_block=semantic_content.merged_block)
            ))
            is_first_date = False
            continue
        tei_child_parent.extend(
            context.get_tei_child_elements_for_semantic_content(semantic_content)
        )
    return tei_ref.element


def get_tei_raw_reference_list_element(
    semantic_reference_list: SemanticReferenceList,
    context: TeiElementFactoryContext
) -> etree.ElementBase:
    tei_reference_list = TeiElementBuilder(TEI_E('listBibl'))
    for semantic_content in semantic_reference_list:
        if isinstance(semantic_content, SemanticRawReference):
            tei_reference_list.append(
                _get_tei_raw_reference_element(semantic_content, context=context)
            )
            continue
        if isinstance(semantic_content, SemanticReference):
            tei_reference_list.append(
                get_tei_reference_element(semantic_content, context=context)
            )
            continue
        tei_reference_list.extend(context.get_tei_child_elements_for_semantic_content(
            semantic_content
        ))
    return tei_reference_list.element


class SemanticReferenceListTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticReferenceList)
        semantic_reference_list = semantic_content
        return get_tei_raw_reference_list_element(
            semantic_reference_list=semantic_reference_list,
            context=context
        )



from flask import Blueprint


class IndexBlueprint(Blueprint):
    def __init__(self):
        super().__init__('index', __name__)
        self.route('/')(self.index)

    def index(self):
        return "ScienceBeam Parser"



import logging

from sciencebeam_parser.document.tei.common import TEI_E
from sciencebeam_parser.models.training_data import (
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/AffiliationAddressParser.java

ROOT_TRAINING_XML_ELEMENT_PATH = [
    'teiHeader', 'fileDesc', 'sourceDesc', 'biblStruct', 'analytic', 'author',
    'affiliation'
]

TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['marker'],
    '<institution>': ROOT_TRAINING_XML_ELEMENT_PATH + ['orgName[@type="institution"]'],
    '<department>': ROOT_TRAINING_XML_ELEMENT_PATH + ['orgName[@type="department"]'],
    '<laboratory>': ROOT_TRAINING_XML_ELEMENT_PATH + ['orgName[@type="laboratory"]'],
    '<addrLine>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address', 'addrLine'],
    '<postCode>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address', 'postCode'],
    '<postBox>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address', 'postBox'],
    '<region>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address', 'region'],
    '<settlement>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address', 'settlement'],
    '<country>': ROOT_TRAINING_XML_ELEMENT_PATH + ['address', 'country']
}


class AffiliationAddressTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.affiliation.tei.xml'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            element_maker=TEI_E,
            default_tei_filename_suffix=(
                AffiliationAddressTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=None,
            default_tei_sub_directory='affiliation-address/corpus'
        )


class AffiliationAddressTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL
            ),
            use_tei_namespace=True
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Sequence

import PIL.Image

from sciencebeam_parser.utils.bounding_box import BoundingBox
from sciencebeam_parser.utils.lazy import LazyLoaded, Preloadable


class ComputerVisionModelInstance(ABC):
    @abstractmethod
    def get_bounding_box(self) -> BoundingBox:
        pass

    @abstractmethod
    def get_type_name(self) -> str:
        pass


@dataclass
class SimpleComputerVisionModelInstance(ComputerVisionModelInstance):
    bounding_box: BoundingBox
    type_name: str

    def get_bounding_box(self) -> BoundingBox:
        return self.bounding_box

    def get_type_name(self) -> str:
        return self.type_name


class ComputerVisionModelResult(ABC):
    @abstractmethod
    def get_instances_by_type_names(
        self,
        type_names: Sequence[str]
    ) -> Sequence[ComputerVisionModelInstance]:
        pass

    def get_instances_by_type_name(
        self,
        type_name: str
    ) -> Sequence[ComputerVisionModelInstance]:
        return self.get_instances_by_type_names([type_name])


class ComputerVisionModel(ABC, Preloadable):
    @abstractmethod
    def predict_single(self, image: PIL.Image.Image) -> ComputerVisionModelResult:
        pass


T_ComputerVisionModelFactory = Callable[[], ComputerVisionModel]


class LazyComputerVisionModel(ComputerVisionModel):
    def __init__(self, factory: T_ComputerVisionModelFactory) -> None:
        super().__init__()
        self._lazy_model = LazyLoaded[ComputerVisionModel](factory)

    def __repr__(self) -> str:
        return '%s(factory=%r, loaded=%r)' % (
            type(self).__name__, self._lazy_model.factory, self._lazy_model.is_loaded
        )

    @property
    def cv_model(self) -> ComputerVisionModel:
        return self._lazy_model.get()

    def preload(self):
        self.cv_model.preload()

    def predict_single(self, image: PIL.Image.Image) -> ComputerVisionModelResult:
        return self.cv_model.predict_single(image)



import logging
from typing import Optional, List, Tuple

import tensorflow as tf

from sciencebeam_trainer_delft.embedding.manager import EmbeddingManager
from sciencebeam_trainer_delft.sequence_labelling.wrapper import (
    DEFAULT_EMBEDDINGS_PATH,
    Sequence
)

from sciencebeam_parser.app.context import AppContext
from sciencebeam_parser.models.model_impl import ModelImpl
from sciencebeam_parser.utils.lazy import LazyLoaded


LOGGER = logging.getLogger(__name__)


class SeparateSessionSequenceWrapper(Sequence):
    def __init__(self, *args, **kwargs):
        self._graph = tf.Graph()
        self._session = tf.Session(graph=self._graph)
        super().__init__(*args, **kwargs)

    def load_from(self, *args, **kwargs):
        with self._graph.as_default():
            with self._session.as_default():
                return super().load_from(*args, **kwargs)

    def tag(self, *args, **kwargs):
        with self._graph.as_default():
            with self._session.as_default():
                return super().tag(*args, **kwargs)


class DelftModelImpl(ModelImpl):
    def __init__(self, model_url: str, app_context: AppContext):
        self.model_url = model_url
        self.app_context = app_context
        self._lazy_model = LazyLoaded[Sequence](self._load_model)

    def __repr__(self) -> str:
        return '%s(%r, loaded=%r)' % (
            type(self).__name__, self.model_url, self._lazy_model.is_loaded
        )

    def _load_model(self) -> Sequence:
        embedding_registry_path = DEFAULT_EMBEDDINGS_PATH
        embedding_manager = EmbeddingManager(
            path=embedding_registry_path,
            download_manager=self.app_context.download_manager
        )
        model = SeparateSessionSequenceWrapper(
            'dummy-model',
            embedding_manager=embedding_manager
        )
        model.load_from(self.model_url)
        LOGGER.info('loaded delft model: %r', self.model_url)
        return model

    @property
    def model(self) -> Sequence:
        return self._lazy_model.get()

    def preload(self):
        self._lazy_model.get()

    def predict_labels(
        self,
        texts: List[List[str]],
        features: List[List[List[str]]],
        output_format: Optional[str] = None
    ) -> List[List[Tuple[str, str]]]:
        model = self.model
        return model.tag(texts, features=features, output_format=output_format)



import logging
from typing import Iterable, Set, Union
from sciencebeam_parser.document.semantic_document import SemanticAuthor

from sciencebeam_parser.models.data import LayoutModelData
from sciencebeam_parser.models.model import (
    LabeledLayoutToken,
    iter_entity_layout_blocks_for_labeled_layout_tokens
)
from sciencebeam_parser.models.name.extract import NameSemanticExtractor
from sciencebeam_parser.models.training_data import (
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser,
    ExtractInstruction,
    ResetExtractInstruction,
    get_model_data_label
)


LOGGER = logging.getLogger(__name__)


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/AuthorParser.java

ROOT_TRAINING_XML_ELEMENT_PATH = [
    'teiHeader', 'fileDesc', 'sourceDesc', 'biblStruct', 'analytic', 'author', 'persName'
]

TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<marker>': ROOT_TRAINING_XML_ELEMENT_PATH + ['marker'],
    '<title>': ROOT_TRAINING_XML_ELEMENT_PATH + ['roleName'],
    '<forename>': ROOT_TRAINING_XML_ELEMENT_PATH + ['forename'],
    '<middlename>': ROOT_TRAINING_XML_ELEMENT_PATH + ['middlename'],
    '<surname>': ROOT_TRAINING_XML_ELEMENT_PATH + ['surname'],
    '<suffix>': ROOT_TRAINING_XML_ELEMENT_PATH + ['suffix']
}


def iter_model_data_with_reset_instruction_iterable(
    model_data_or_instruction_iterable: Iterable[Union[LayoutModelData, ExtractInstruction]]
) -> Iterable[Union[LayoutModelData, ExtractInstruction]]:
    # using extractor to re-use logic to split author names
    # here we will split on the first token of the extracted semantic content
    extractor = NameSemanticExtractor()
    model_data_or_instruction_list = list(
        model_data_or_instruction_iterable
    )
    entity_tokens = iter_entity_layout_blocks_for_labeled_layout_tokens([
        LabeledLayoutToken(
            label=get_model_data_label(model_data) or '',
            layout_token=model_data.layout_token
        )
        for model_data in model_data_or_instruction_list
        if (
            isinstance(model_data, LayoutModelData)
            and model_data.layout_token is not None
        )
    ])
    LOGGER.debug('entity_tokens: %r', entity_tokens)
    reset_token_ids: Set[int] = set()
    for index, semantic_content in enumerate(extractor.iter_semantic_content_for_entity_blocks(
        entity_tokens=entity_tokens,
        name_type=SemanticAuthor
    )):
        if index == 0:
            continue
        for semantic_token in semantic_content.iter_tokens():
            reset_token_ids.add(id(semantic_token))
            break
    for model_data_or_instruction in model_data_or_instruction_list:
        if isinstance(model_data_or_instruction, LayoutModelData):
            model_data = model_data_or_instruction
            if id(model_data.layout_token) in reset_token_ids:
                yield ResetExtractInstruction(
                    ROOT_TRAINING_XML_ELEMENT_PATH[:-1]
                )
        yield model_data_or_instruction


class NameTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.authors.tei.xml'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            use_tei_namespace=True,
            root_tag='TEI',
            default_tei_filename_suffix=(
                NameTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=None
        )

    def iter_model_data_or_instruction_for_model_data_iterable(
        self,
        model_data_iterable: Iterable[LayoutModelData]
    ) -> Iterable[Union[LayoutModelData, ExtractInstruction]]:
        parent_model_data_or_instruction_iterable = (
            super().iter_model_data_or_instruction_for_model_data_iterable(
                model_data_iterable
            )
        )
        return iter_model_data_with_reset_instruction_iterable(
            parent_model_data_or_instruction_iterable
        )


class NameTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH[:-1],
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL
            ),
            use_tei_namespace=True
        )



import atexit
import logging
import signal
import subprocess
import time
from functools import partial
from threading import Timer, Thread
from typing import Iterable, Optional, Sequence, Union


LOGGER = logging.getLogger(__name__)


class ChildProcessReturnCodeError(ChildProcessError):
    def __init__(self, *args, returncode: int, process=None):
        super().__init__(*args)
        self.returncode = returncode
        self.process = process


class ChildProcessTimeoutError(ChildProcessReturnCodeError):
    pass


class BackgroundProcess:
    def __init__(self, process: subprocess.Popen):
        self.process = process
        self._stopped_by_timeout = False
        self._created_time = time.monotonic()
        self._returncode: Optional[int] = None

    def __repr__(self) -> str:
        return (
            '{type_name}('
            'pid={self.process.pid}'
            ', returncode={self.process.returncode}'
            ')'
        ).format(
            type_name=type(self).__name__,
            self=self
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, value, traceback):
        self.process.__exit__(exc_type, value, traceback)

    def is_running(self) -> bool:
        self.process.poll()
        return self.returncode is None

    @property
    def returncode(self) -> Optional[int]:
        if self.process.returncode != self._returncode:
            self._returncode = self.process.returncode
            LOGGER.debug('process(pid=%s).returncode: %s', self.process.pid, self._returncode)
        return self._returncode

    @property
    def pid(self) -> Optional[int]:
        return self.process.pid

    def send_signal(self, sig: int) -> None:
        if self.process.returncode is not None:
            LOGGER.debug(
                'not sending signal %r, process has already stopped: %s',
                sig, self.process.pid
            )
            return
        LOGGER.info('sending %s to process %s', sig, self.process.pid)
        self.process.send_signal(sig)

    def terminate(self) -> None:
        self.send_signal(signal.SIGINT)

    def kill(self) -> None:
        self.send_signal(signal.SIGKILL)

    def kill_if_runing(self) -> None:
        if not self.is_running():
            return
        self.kill()

    def wait(self) -> int:
        self.process.wait()
        # Note: not using the return type in order to use the logging logic of returncode function
        returncode = self.returncode
        assert returncode is not None
        return returncode

    def get_uptime(self) -> float:
        return time.monotonic() - self._created_time

    def stop(self, wait: bool = True, kill_timeout: int = 60) -> None:
        self.terminate()
        if kill_timeout:
            Timer(kill_timeout, self.kill_if_runing).start()
        if wait:
            LOGGER.info('waiting for process(pid=%s) to stop', self.process.pid)
            self.wait()
            LOGGER.info(
                'process(pid=%s) has stopped with returncode: %s',
                self.process.pid, self.returncode
            )

    def stop_due_to_timeout(self, **kwargs) -> None:
        LOGGER.info('process timeout, stopping: %s', self.process.pid)
        self._stopped_by_timeout = True
        self.stop(**kwargs)

    def is_stopped_by_timeout(self) -> bool:
        return self._stopped_by_timeout

    def check_returncode(self) -> None:
        returncode = self.process.returncode
        if returncode is None:
            return
        if self.is_stopped_by_timeout():
            LOGGER.debug('process stopped by timeout, return code: %s', returncode)
            raise ChildProcessTimeoutError(
                'process stopped by timeout, return code: %s' % returncode,
                returncode=returncode
            )
        if returncode != 0:
            LOGGER.debug('process failed with return code: %s', returncode)
            raise ChildProcessReturnCodeError(
                'process failed with return code: %s' % returncode,
                returncode=returncode
            )

    def stop_if_running(self, **kwargs) -> None:
        if not self.is_running():
            return
        self.stop(**kwargs)


def stream_lines_to_logger(
    lines: Iterable[str],
    logger: logging.Logger,
    prefix: str = ''
):
    for line in lines:
        line = line.strip()
        if line:
            logger.info('%s%s', prefix, line)


def exec_with_logging(
    command: Union[str, Sequence[str]],
    logging_prefix: Optional[str] = None,
    process_timeout: Optional[float] = None,
    daemon: bool = False,
    check_returncode: bool = True,
    **kwargs
) -> BackgroundProcess:
    p = BackgroundProcess(subprocess.Popen(  # pylint: disable=consider-using-with
        command,
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
        **kwargs
    ))
    if logging_prefix is None:
        logging_prefix = 'process'
    logging_prefix += '[pid:%s]: ' % p.process.pid
    if not daemon:
        timer = None
        if process_timeout:
            timer = Timer(process_timeout, p.stop_due_to_timeout)
            timer.start()
        assert p.process.stdout is not None
        stream_lines_to_logger(p.process.stdout, LOGGER, logging_prefix)
        p.wait()
        if timer:
            timer.cancel()
        if check_returncode:
            p.check_returncode()
        return p
    t = Thread(target=partial(
        stream_lines_to_logger,
        lines=p.process.stdout,
        logger=LOGGER,
        prefix=logging_prefix
    ))
    t.daemon = True
    t.start()
    return p


class CommandRestartableBackgroundProcess:
    def __init__(
        self,
        command: Union[str, Sequence[str]],
        name: str = None,
        logging_prefix: str = None,
        stop_at_exit: bool = False
    ):
        self.command = command
        self.name = name
        self.logging_prefix = logging_prefix
        self.process: Optional[BackgroundProcess] = None
        self.stop_at_exit = stop_at_exit
        self._atexit_registered = False

    def stop(self, wait: bool = True) -> None:
        if self.process:
            self.process.stop(wait=wait)

    def stop_if_running(self, wait: bool = True, **kwargs) -> None:
        if self.process:
            self.process.stop_if_running(wait=wait, **kwargs)

    def start(self, stop: bool = True) -> None:
        if stop:
            self.stop_if_running(wait=True)
        if self.stop_at_exit and not self._atexit_registered:
            atexit.register(self.stop_if_running)
            self._atexit_registered = True
        LOGGER.info('starting %s', self.name)
        LOGGER.debug('running background command: %s', self.command)
        self.process = exec_with_logging(
            self.command,
            logging_prefix=self.logging_prefix or self.name,
            daemon=True
        )

    def is_running(self) -> bool:
        return self.process is not None and self.process.is_running()

    def get_uptime(self) -> float:
        assert self.process is not None
        return self.process.get_uptime()

    def start_if_not_running(self) -> None:
        if not self.is_running():
            if self.process:
                LOGGER.info('process has stopped, restarting: %s', self.process.pid)
            self.start()



from abc import ABC, abstractmethod
import logging
import re
from typing import Iterable, Mapping, Optional, Tuple

from sciencebeam_parser.document.semantic_document import (
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticNote
)
from sciencebeam_parser.document.layout_document import EMPTY_BLOCK, LayoutBlock, LayoutTokensText


LOGGER = logging.getLogger(__name__)


class ModelSemanticExtractor(ABC):
    @abstractmethod
    def iter_semantic_content_for_entity_blocks(
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        pass


def get_regex_cleaned_layout_block_with_prefix_suffix(
    layout_block: LayoutBlock,
    regex_pattern: Optional[str]
) -> Tuple[LayoutBlock, LayoutBlock, LayoutBlock]:
    if not layout_block or not layout_block.lines or not regex_pattern:
        return EMPTY_BLOCK, layout_block, EMPTY_BLOCK
    layout_tokens_text = LayoutTokensText(layout_block)
    text = str(layout_tokens_text)
    m = re.match(regex_pattern, text, re.IGNORECASE)
    if not m:
        LOGGER.debug('text does not match regex: %r', text)
        return EMPTY_BLOCK, layout_block, EMPTY_BLOCK
    start = m.start(1)
    end = m.end(1)
    LOGGER.debug('start: %d, end: %d, len: %d (text: %r)', start, end, len(text), text)
    return (
        LayoutBlock.for_tokens(list(
            layout_tokens_text.iter_layout_tokens_between(0, start)
        )),
        LayoutBlock.for_tokens(list(
            layout_tokens_text.iter_layout_tokens_between(start, end)
        )),
        LayoutBlock.for_tokens(list(
            layout_tokens_text.iter_layout_tokens_between(end, len(text))
        ))
    )


class SimpleModelSemanticExtractor(ModelSemanticExtractor):
    def __init__(
        self,
        semantic_content_class_by_tag: Optional[
            Mapping[str, SemanticContentFactoryProtocol]
        ] = None
    ):
        super().__init__()
        self.semantic_content_class_by_tag = semantic_content_class_by_tag or {}

    def get_semantic_content_for_entity_name(
        self,
        name: str,
        layout_block: LayoutBlock
    ) -> SemanticContentWrapper:
        semantic_content_class = self.semantic_content_class_by_tag.get(name)
        if semantic_content_class:
            return semantic_content_class(layout_block=layout_block)
        return SemanticNote(
            layout_block=layout_block,
            note_type=name
        )



from abc import ABC, abstractmethod
from typing import Callable, List, Optional, Tuple

from sciencebeam_parser.utils.lazy import Preloadable


class ModelImpl(ABC, Preloadable):
    @abstractmethod
    def predict_labels(
        self,
        texts: List[List[str]],
        features: List[List[List[str]]],
        output_format: Optional[str] = None
    ) -> List[List[Tuple[str, str]]]:
        pass


T_ModelImplFactory = Callable[[], ModelImpl]



import re
from typing import Iterable, List


# delimters mostly copied from:
# https://github.com/kermitt2/delft/blob/v0.2.6/delft/utilities/Tokenizer.py
# and:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/utilities/TextUtilities.java#L773-L948
# added: `@`, `#`, `\u2020`, ...
_COMMON_AFF_MARKERS = '\u2020\u2021\u00A7\u00B6\u204B\u01C2'
DELIMITERS = (
    "\n\r\t\f\u00A0([ •*,:;?.!/#)-−–‐\"“”‘’'`$]*\u2666\u2665\u2663\u2660\u00A0@"
    + _COMMON_AFF_MARKERS
)
DELIMITERS_REGEX = r'(' + r'|'.join(map(re.escape, DELIMITERS)) + r'|\s)'


def iter_tokenized_tokens(text: str, keep_whitespace: bool = False) -> Iterable[str]:
    for token in re.split(DELIMITERS_REGEX, text):
        if not keep_whitespace and not token.strip():
            continue
        yield token


def get_tokenized_tokens(text: str, **kwargs) -> List[str]:
    return list(iter_tokenized_tokens(text, **kwargs))



import dataclasses
import logging
from dataclasses import dataclass
from typing import (
    Any,
    Iterable,
    Optional,
    Type,
    TypeVar
)

from sciencebeam_parser.app.context import AppContext
from sciencebeam_parser.config.config import AppConfig
from sciencebeam_parser.models.model import Model
from sciencebeam_parser.models.model_impl_factory import get_model_impl_factory_for_config
from sciencebeam_parser.cv_models.cv_model_factory import get_lazy_cv_model_for_config
from sciencebeam_parser.cv_models.cv_model import ComputerVisionModel
from sciencebeam_parser.ocr_models.ocr_model_factory import get_lazy_ocr_model_for_config
from sciencebeam_parser.ocr_models.ocr_model import OpticalCharacterRecognitionModel

from sciencebeam_parser.models.segmentation.model import SegmentationModel
from sciencebeam_parser.models.header.model import HeaderModel
from sciencebeam_parser.models.name.model import NameModel
from sciencebeam_parser.models.affiliation_address.model import AffiliationAddressModel
from sciencebeam_parser.models.fulltext.model import FullTextModel
from sciencebeam_parser.models.figure.model import FigureModel
from sciencebeam_parser.models.table.model import TableModel
from sciencebeam_parser.models.reference_segmenter.model import ReferenceSegmenterModel
from sciencebeam_parser.models.citation.model import CitationModel
from sciencebeam_parser.processors.fulltext.config import FullTextProcessorConfig


LOGGER = logging.getLogger(__name__)


@dataclass
class FullTextModels:
    segmentation_model: SegmentationModel
    header_model: HeaderModel
    name_header_model: NameModel
    name_citation_model: NameModel
    affiliation_address_model: AffiliationAddressModel
    fulltext_model: FullTextModel
    figure_model: FigureModel
    table_model: TableModel
    reference_segmenter_model: ReferenceSegmenterModel
    citation_model: CitationModel
    cv_model: Optional[ComputerVisionModel] = None
    ocr_model: Optional[OpticalCharacterRecognitionModel] = None

    def _iter_models(self) -> Iterable[Any]:
        for field in dataclasses.fields(self):
            model = getattr(self, field.name)
            if model is not None:
                yield model

    def preload(self):
        LOGGER.info('preloading models...')
        models = list(self._iter_models())
        for model_index, model in enumerate(models):
            LOGGER.info('preloading: %d/%d: %r', 1 + model_index, len(models), model)
            try:
                model.preload()
            except AttributeError:
                pass
        LOGGER.info('preloaded models')

    def get_sequence_model_by_name(self, model_name: str) -> Model:
        model = getattr(self, model_name + '_model')
        assert isinstance(model, Model)
        return model


T_Model = TypeVar('T_Model', bound=Model)


def load_model(
    app_config: AppConfig,
    app_context: AppContext,
    model_name: str,
    model_class: Type[T_Model]
) -> T_Model:
    models_config = app_config['models']
    model_config = models_config[model_name]
    model = model_class(
        get_model_impl_factory_for_config(
            model_config,
            app_context=app_context
        ),
        model_config=model_config
    )
    return model


def get_cv_model_for_app_config(
    app_config: AppConfig,
    enabled: bool = True
) -> Optional[ComputerVisionModel]:
    cv_model_config = app_config.get('cv_models', {}).get('default')
    if enabled and cv_model_config:
        return get_lazy_cv_model_for_config(cv_model_config)
    return None


def get_ocr_model_for_app_config(
    app_config: AppConfig,
    enabled: bool = True
) -> Optional[OpticalCharacterRecognitionModel]:
    ocr_model_config = app_config.get('ocr_models', {}).get('default')
    if enabled and ocr_model_config:
        return get_lazy_ocr_model_for_config(ocr_model_config)
    return None


def load_models(
    app_config: AppConfig,
    app_context: AppContext,
    fulltext_processor_config: FullTextProcessorConfig
) -> FullTextModels:
    segmentation_model = load_model(
        app_config, app_context, 'segmentation', SegmentationModel
    )
    header_model = load_model(
        app_config, app_context, 'header', HeaderModel
    )
    name_header_model = load_model(
        app_config, app_context, 'name_header', NameModel
    )
    name_citation_model = load_model(
        app_config, app_context, 'name_citation', NameModel
    )
    affiliation_address_model = load_model(
        app_config, app_context, 'affiliation_address', AffiliationAddressModel
    )
    fulltext_model = load_model(
        app_config, app_context, 'fulltext', FullTextModel
    )
    figure_model = load_model(
        app_config, app_context, 'figure', FigureModel
    )
    table_model = load_model(
        app_config, app_context, 'table', TableModel
    )
    reference_segmenter_model = load_model(
        app_config, app_context, 'reference_segmenter', ReferenceSegmenterModel
    )
    citation_model = load_model(
        app_config, app_context, 'citation', CitationModel
    )
    cv_model = get_cv_model_for_app_config(
        app_config,
        enabled=fulltext_processor_config.use_cv_model
    )
    ocr_model = get_ocr_model_for_app_config(
        app_config,
        enabled=fulltext_processor_config.use_ocr_model
    )
    return FullTextModels(
        segmentation_model=segmentation_model,
        header_model=header_model,
        name_header_model=name_header_model,
        name_citation_model=name_citation_model,
        affiliation_address_model=affiliation_address_model,
        fulltext_model=fulltext_model,
        figure_model=figure_model,
        table_model=table_model,
        reference_segmenter_model=reference_segmenter_model,
        citation_model=citation_model,
        cv_model=cv_model,
        ocr_model=ocr_model
    )



import logging
import re
from typing import Iterable, Mapping, Optional, Set, Tuple, Type, Union

from sciencebeam_parser.utils.misc import iter_ids
from sciencebeam_parser.document.semantic_document import (
    SemanticContentFactoryProtocol,
    SemanticContentWrapper,
    SemanticDate,
    SemanticExternalIdentifier,
    SemanticExternalIdentifierTypes,
    SemanticExternalUrl,
    SemanticInvalidReference,
    SemanticIssue,
    SemanticJournal,
    SemanticLocation,
    SemanticPageRange,
    SemanticPublisher,
    SemanticRawAuthors,
    SemanticRawEditors,
    SemanticRawReference,
    SemanticRawReferenceText,
    SemanticReference,
    SemanticTitle,
    SemanticVolume
)
from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.models.extract import SimpleModelSemanticExtractor


LOGGER = logging.getLogger(__name__)


# https://en.wikipedia.org/wiki/Digital_Object_Identifier
# https://www.doi.org/doi_handbook/2_Numbering.html
DOI_PATTERN = r'\b(10\.\d{4,}(?:\.\d{1,})*/.+)'

# copied and adapted from:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/utilities/TextUtilities.java#L66
PMID_PATTERN = r"(?:(?:PMID)|(?:Pub(?:\s)?Med(?:\s)?(?:ID)?))(?:\s)?(?:\:)?(?:\s)*(\d{1,8})"

PMCID_PATTERN = r"(?:PMC)(\d{1,})"

# copied and adapted from:
# https://github.com/kermitt2/grobid/blob/0.6.2/grobid-core/src/main/java/org/grobid/core/utilities/TextUtilities.java#L62-L63
ARXIV_PATTERN = (
    r"(?:arXiv\s?(?:\.org)?\s?\:\s?(\d{4}\s?\.\s?\d{4,5}(?:v\d+)?))"
    r"|(?:arXiv\s?(?:\.org)?\s?\:\s?([ a-zA-Z\-\.]*\s?/\s?\d{7}(?:v\d+)?))"
)

# https://en.wikipedia.org/wiki/Publisher_Item_Identifier
PII_PATTERN = r'\b([S,B]\W*(?:[0-9xX]\W*){15,}[0-9xX])'


SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG: Mapping[str, SemanticContentFactoryProtocol] = {
    '<author>': SemanticRawAuthors,
    '<editor>': SemanticRawEditors,
    '<title>': SemanticTitle,
    '<journal>': SemanticJournal,
    '<volume>': SemanticVolume,
    '<issue>': SemanticIssue,
    '<publisher>': SemanticPublisher,
    '<location>': SemanticLocation
}


VALID_REFERENCE_TYPES: Set[Type[SemanticContentWrapper]] = {
    SemanticTitle,
    SemanticJournal,
    SemanticRawAuthors,
    SemanticRawEditors,
    SemanticExternalIdentifier,
    SemanticExternalUrl
}


def parse_page_range(layout_block: LayoutBlock) -> SemanticPageRange:
    page_range_text = layout_block.text
    page_parts = page_range_text.split('-')
    if len(page_parts) == 2:
        from_page = page_parts[0].strip()
        to_page = page_parts[1].strip()
        if to_page and len(to_page) < len(from_page):
            to_page = from_page[:-(len(to_page))] + to_page
        return SemanticPageRange(
            layout_block=layout_block,
            from_page=from_page,
            to_page=to_page
        )
    return SemanticPageRange(layout_block=layout_block)


def parse_web(layout_block: LayoutBlock) -> Union[SemanticExternalUrl, SemanticExternalIdentifier]:
    value = re.sub(r'\s', '', layout_block.text)
    m = re.search(DOI_PATTERN, value)
    if m:
        return SemanticExternalIdentifier(
            layout_block=layout_block,
            value=m.group(1),
            external_identifier_type=SemanticExternalIdentifierTypes.DOI
        )
    return SemanticExternalUrl(
        layout_block=layout_block,
        value=value
    )


def get_detected_external_identifier_type_and_value_for_text(
    text: str
) -> Tuple[Optional[str], str]:
    value = re.sub(r'\s', '', text)
    m = re.search(DOI_PATTERN, value)
    if m:
        value = m.group(1)
        return SemanticExternalIdentifierTypes.DOI, value
    m = re.search(PMCID_PATTERN, value)
    if m:
        value = 'PMC' + m.group(1)
        return SemanticExternalIdentifierTypes.PMCID, value
    m = re.search(ARXIV_PATTERN, value)
    if m:
        value = m.group(1) or m.group(2)
        return SemanticExternalIdentifierTypes.ARXIV, value
    m = re.match(PMID_PATTERN, value)
    if m:
        value = m.group(1)
        return SemanticExternalIdentifierTypes.PMID, value
    m = re.search(PII_PATTERN, value)
    if m:
        value = m.group(1)
        return SemanticExternalIdentifierTypes.PII, value
    return None, value


def get_detected_external_identifier_type_for_text(text: str) -> Optional[str]:
    external_identifier_type, _ = get_detected_external_identifier_type_and_value_for_text(
        text
    )
    return external_identifier_type


def parse_pubnum(layout_block: LayoutBlock) -> SemanticExternalIdentifier:
    external_identifier_type, value = get_detected_external_identifier_type_and_value_for_text(
        layout_block.text
    )
    return SemanticExternalIdentifier(
        layout_block=layout_block,
        value=value,
        external_identifier_type=external_identifier_type
    )


def parse_date(layout_block: LayoutBlock) -> SemanticDate:
    value = re.sub(r'\s', '', layout_block.text)
    year: Optional[int] = None
    m = re.search(r'(\d{4})', value)
    if m:
        year = int(m.group(1))
    return SemanticDate(
        layout_block=layout_block,
        year=year
    )


def is_reference_valid(ref: SemanticReference) -> bool:
    for semantic_content in ref:
        if type(semantic_content) in VALID_REFERENCE_TYPES:
            return True
    return False


def get_invalid_reference(ref: SemanticReference) -> SemanticInvalidReference:
    return SemanticInvalidReference(
        mixed_content=[
            semantic_content
            for semantic_content in ref
            if not isinstance(semantic_content, SemanticRawReferenceText)
        ]
    )


class CitationSemanticExtractor(SimpleModelSemanticExtractor):
    def __init__(self):
        super().__init__(semantic_content_class_by_tag=SIMPLE_SEMANTIC_CONTENT_CLASS_BY_TAG)

    def get_semantic_content_for_entity_name(  # pylint: disable=too-many-return-statements
        self,
        name: str,
        layout_block: LayoutBlock
    ) -> SemanticContentWrapper:
        if name == '<pages>':
            return parse_page_range(layout_block)
        if name == '<web>':
            return parse_web(layout_block)
        if name == '<pubnum>':
            return parse_pubnum(layout_block)
        if name == '<date>':
            return parse_date(layout_block)
        return super().get_semantic_content_for_entity_name(name, layout_block)

    def iter_semantic_content_for_entity_blocks(  # pylint: disable=arguments-differ
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        semantic_raw_reference: Optional[SemanticRawReference] = None,
        **kwargs
    ) -> Iterable[SemanticContentWrapper]:
        entity_tokens = list(entity_tokens)
        LOGGER.debug('entity_tokens: %s', entity_tokens)
        ids_iterator = iter(iter_ids('b'))
        ref: Optional[SemanticReference] = None
        for name, layout_block in entity_tokens:
            if not ref:
                ref = SemanticReference()
                if semantic_raw_reference:
                    ref.content_id = semantic_raw_reference.content_id
                    for semantic_content in semantic_raw_reference:
                        ref.add_content(semantic_content)
                if not ref.content_id:
                    ref.content_id = next(ids_iterator, '?')
            semantic_content = self.get_semantic_content_for_entity_name(
                name, layout_block=layout_block
            )
            ref.add_content(semantic_content)
        if ref and not is_reference_valid(ref):
            yield get_invalid_reference(ref)
        elif ref:
            yield ref



from abc import ABC, abstractmethod
import argparse
import logging
import os
from dataclasses import dataclass, field
from typing import Dict, Iterable, List, NamedTuple, Optional, Sequence

from lxml import etree

from sciencebeam_trainer_delft.utils.io import (
    auto_download_input_file
)

from sciencebeam_parser.utils.io import glob, makedirs, write_bytes, write_text

from sciencebeam_parser.document.layout_document import LayoutDocument
from sciencebeam_parser.document.semantic_document import (
    SemanticMixedContentWrapper,
    SemanticRawAffiliationAddress,
    SemanticRawAuthors,
    SemanticRawFigure,
    SemanticRawReference,
    SemanticRawReferenceText,
    SemanticRawTable
)
from sciencebeam_parser.models.data import (
    DocumentFeaturesContext,
    LabeledLayoutModelData,
    LayoutModelData
)
from sciencebeam_parser.models.model import (
    LabeledLayoutToken,
    LayoutDocumentLabelResult,
    LayoutModelLabel,
    Model,
    iter_data_lines_for_model_data_iterables,
    iter_labeled_layout_token_for_layout_model_label
)
from sciencebeam_parser.models.training_data import TeiTrainingDataGenerator
from sciencebeam_parser.processors.fulltext.models import FullTextModels
from sciencebeam_parser.resources.default_config import DEFAULT_CONFIG_FILE
from sciencebeam_parser.config.config import AppConfig
from sciencebeam_parser.app.parser import ScienceBeamParser
from sciencebeam_parser.utils.media_types import MediaTypes


LOGGER = logging.getLogger(__name__)


@dataclass
class ModelResultCache:
    model_data_lists_by_key_map: Dict[
        str, Sequence[Sequence[LabeledLayoutModelData]]
    ] = field(default_factory=dict)


def parse_args(argv: Optional[List[str]] = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        'ScienceBeam Parser: Generate Training Data'
    )
    parser.add_argument(
        '--source-path',
        type=str,
        required=True
    )
    parser.add_argument(
        '--output-path',
        type=str,
        required=True
    )
    parser.add_argument(
        '--limit',
        type=int,
        required=False
    )
    parser.add_argument(
        '--use-model',
        action='store_true',
        help='Use configured models to pre-annotate training data'
    )
    parser.add_argument(
        '--use-directory-structure',
        action='store_true',
        help='Output training data to a directory structure'
    )
    parser.add_argument(
        '--gzip',
        action='store_true',
        help='Enable gzip compression for output files (with .gz suffix)'
    )
    parser.add_argument(
        '--debug',
        action='store_true',
        help='Enable debug logging'
    )
    return parser.parse_args(argv)


def get_labeled_model_data_list_list(
    model_data_list_list: Sequence[Sequence[LayoutModelData]],
    model: Model
) -> Sequence[Sequence[LabeledLayoutModelData]]:
    return list(
        model.iter_labeled_model_data_list_for_model_data_list_iterable(
            model_data_list_list
        )
    )


def get_labeled_model_data_list(
    model_data_list: Sequence[LayoutModelData],
    model: Model
) -> Sequence[LabeledLayoutModelData]:
    return get_labeled_model_data_list_list(
        [model_data_list],
        model=model
    )[0]


def get_labeled_model_data_list_for_layout_document(
    layout_document: LayoutDocument,
    model: Model,
    document_features_context: DocumentFeaturesContext
) -> Sequence[LabeledLayoutModelData]:
    data_generator = model.get_data_generator(
        document_features_context=document_features_context
    )
    model_data_list: Sequence[LayoutModelData] = list(
        data_generator.iter_model_data_for_layout_document(layout_document)
    )
    return get_labeled_model_data_list(
        model_data_list,
        model=model
    )


def get_layout_model_label_for_labeled_model_data(
    labeled_model_data: LabeledLayoutModelData
) -> LayoutModelLabel:
    return LayoutModelLabel(
        label=labeled_model_data.label or '',
        label_token_text=labeled_model_data.label_token_text,
        layout_line=labeled_model_data.layout_line,
        layout_token=labeled_model_data.layout_token
    )


def iter_layout_model_label_for_labeled_model_data_list(
    labeled_model_data_iterable: Iterable[LabeledLayoutModelData],
) -> Iterable[LayoutModelLabel]:
    return (
        get_layout_model_label_for_labeled_model_data(labeled_model_data)
        for labeled_model_data in labeled_model_data_iterable
    )


def get_layout_document_label_result_for_labeled_model_data_list(
    labeled_model_data_iterable: Iterable[LabeledLayoutModelData],
    layout_document: LayoutDocument
) -> LayoutDocumentLabelResult:
    return LayoutDocumentLabelResult(
        layout_document=layout_document,
        layout_model_label_iterable=iter_layout_model_label_for_labeled_model_data_list(
            labeled_model_data_iterable
        )
    )


class TrainingDataDocumentContext(NamedTuple):
    output_path: str
    source_filename: str
    document_features_context: DocumentFeaturesContext
    fulltext_models: FullTextModels
    use_model: bool
    use_directory_structure: bool
    model_result_cache: ModelResultCache
    gzip_enabled: bool

    @property
    def source_name(self) -> str:
        source_basename = os.path.basename(self.source_filename)
        return os.path.splitext(source_basename)[0]


def iter_unlabeled_model_data_list_for_model_and_layout_documents(
    model: Model,
    model_layout_documents: Sequence[LayoutDocument],
    document_context: TrainingDataDocumentContext
) -> Iterable[Sequence[LayoutModelData]]:
    if not model_layout_documents:
        return []
    data_generator = model.get_data_generator(
        document_features_context=document_context.document_features_context
    )
    return [
        list(
            data_generator.iter_model_data_for_layout_document(model_layout_document)
        )
        for model_layout_document in model_layout_documents
    ]


def iter_labeled_model_data_list_for_model_and_layout_documents(
    model: Model,
    model_layout_documents: Sequence[LayoutDocument],
    document_context: TrainingDataDocumentContext
) -> Iterable[Sequence[LabeledLayoutModelData]]:
    if not model_layout_documents:
        return []
    cache_key = f'{type(model).__name__}_{id(model)}'
    LOGGER.debug('cache_key: %r', cache_key)
    model_data_lists = document_context.model_result_cache.model_data_lists_by_key_map.get(
        cache_key
    )
    if model_data_lists is not None:
        return model_data_lists
    unlabeled_model_data_lists = list(
        iter_unlabeled_model_data_list_for_model_and_layout_documents(
            model=model,
            model_layout_documents=model_layout_documents,
            document_context=document_context
        )
    )
    model_data_lists = get_labeled_model_data_list_list(
        unlabeled_model_data_lists,
        model=model
    )
    document_context.model_result_cache.model_data_lists_by_key_map[cache_key] = (
        model_data_lists
    )
    return model_data_lists


def iter_model_data_list_for_model_and_layout_documents(
    model: Model,
    model_layout_documents: Sequence[LayoutDocument],
    document_context: TrainingDataDocumentContext
) -> Iterable[Sequence[LayoutModelData]]:
    if not document_context.use_model:
        return iter_unlabeled_model_data_list_for_model_and_layout_documents(
            model=model,
            model_layout_documents=model_layout_documents,
            document_context=document_context
        )
    return iter_labeled_model_data_list_for_model_and_layout_documents(
        model=model,
        model_layout_documents=model_layout_documents,
        document_context=document_context
    )


def get_labeled_layout_tokens_list_for_model_and_layout_documents(
    model: Model,
    layout_documents: Sequence[LayoutDocument],
    document_context: TrainingDataDocumentContext
) -> Sequence[Sequence[LabeledLayoutToken]]:
    model_data_lists = list(
        iter_labeled_model_data_list_for_model_and_layout_documents(
            model=model,
            model_layout_documents=layout_documents,
            document_context=document_context
        )
    )
    assert len(model_data_lists) == len(layout_documents)
    return [
        list(iter_labeled_layout_token_for_layout_model_label(
            iter_layout_model_label_for_labeled_model_data_list(
                model_data_list
            )
        ))
        for model_data_list in model_data_lists
    ]


def get_labeled_layout_tokens_for_model_and_layout_document(
    model: Model,
    layout_document: LayoutDocument,
    document_context: TrainingDataDocumentContext
) -> Sequence[LabeledLayoutToken]:
    labeled_layout_tokens_list = get_labeled_layout_tokens_list_for_model_and_layout_documents(
        model,
        [layout_document],
        document_context
    )
    assert len(labeled_layout_tokens_list) == 1
    return labeled_layout_tokens_list[0]


def get_segmentation_label_result(
    layout_document: LayoutDocument,
    document_context: TrainingDataDocumentContext
) -> LayoutDocumentLabelResult:
    segmentation_label_model_data_lists = list(
        iter_labeled_model_data_list_for_model_and_layout_documents(
            model=document_context.fulltext_models.segmentation_model,
            model_layout_documents=[layout_document],
            document_context=document_context
        )
    )
    assert len(segmentation_label_model_data_lists) == 1
    LOGGER.debug('segmentation_label_model_data_lists: %r', segmentation_label_model_data_lists)
    return get_layout_document_label_result_for_labeled_model_data_list(
        labeled_model_data_iterable=segmentation_label_model_data_lists[0],
        layout_document=layout_document
    )


class AbstractModelTrainingDataGenerator(ABC):
    def get_pre_file_path_suffix(self) -> str:
        return ''

    def _get_file_path_with_suffix(
        self,
        suffix: Optional[str],
        document_context: TrainingDataDocumentContext,
        sub_directory: Optional[str] = None
    ) -> Optional[str]:
        if not suffix:
            return None
        output_path = document_context.output_path
        if sub_directory and document_context.use_directory_structure:
            output_path = os.path.join(output_path, sub_directory)
        if document_context.gzip_enabled:
            suffix += '.gz'
        return os.path.join(
            output_path,
            document_context.source_name + self.get_pre_file_path_suffix() + suffix
        )

    @abstractmethod
    def get_tei_training_data_generator(
        self,
        document_context: TrainingDataDocumentContext
    ) -> TeiTrainingDataGenerator:
        pass

    @abstractmethod
    def iter_model_data_list(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[Sequence[LayoutModelData]]:
        return []

    def get_default_tei_sub_directory(
        self,
        tei_training_data_generator: TeiTrainingDataGenerator
    ) -> Optional[str]:
        return tei_training_data_generator.get_default_tei_sub_directory()

    def generate_data_for_layout_document(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ):
        tei_training_data_generator = self.get_tei_training_data_generator(document_context)
        tei_file_path = self._get_file_path_with_suffix(
            tei_training_data_generator.get_default_tei_filename_suffix(),
            document_context=document_context,
            sub_directory=self.get_default_tei_sub_directory(tei_training_data_generator)
        )
        data_file_path = self._get_file_path_with_suffix(
            tei_training_data_generator.get_default_data_filename_suffix(),
            document_context=document_context,
            sub_directory=tei_training_data_generator.get_default_data_sub_directory()
        )
        assert tei_file_path
        model_data_list_list = list(self.iter_model_data_list(
            layout_document=layout_document,
            document_context=document_context
        ))
        if not model_data_list_list:
            LOGGER.info('no entities found, skipping (%r)', tei_file_path)
            return
        training_tei_root = (
            tei_training_data_generator
            .get_training_tei_xml_for_multiple_model_data_iterables(
                model_data_list_list
            )
        )
        LOGGER.info('writing training tei to: %r', tei_file_path)
        write_bytes(
            tei_file_path,
            etree.tostring(training_tei_root, pretty_print=True)
        )
        if data_file_path:
            LOGGER.info('writing training raw data to: %r', data_file_path)
            write_text(
                data_file_path,
                '\n'.join(
                    iter_data_lines_for_model_data_iterables(model_data_list_list)
                ),
                encoding='utf-8'
            )


class AbstractDocumentModelTrainingDataGenerator(AbstractModelTrainingDataGenerator):
    @abstractmethod
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        pass

    def get_tei_training_data_generator(
        self,
        document_context: TrainingDataDocumentContext
    ) -> TeiTrainingDataGenerator:
        return self.get_main_model(document_context).get_tei_training_data_generator()

    @abstractmethod
    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        pass

    def iter_model_data_list(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[Sequence[LayoutModelData]]:
        model = self.get_main_model(document_context)
        model_layout_documents = list(self.iter_model_layout_documents(
            layout_document,
            document_context=document_context
        ))
        return iter_model_data_list_for_model_and_layout_documents(
            model=model,
            model_layout_documents=model_layout_documents,
            document_context=document_context
        )


class SegmentationModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.segmentation_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        return [layout_document]


class HeaderModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.header_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        LOGGER.debug('segmentation_label_result: %r', segmentation_label_result)
        header_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<header>'
        ).remove_empty_blocks()
        LOGGER.debug('header_layout_document: %r', header_layout_document)
        if not header_layout_document.pages:
            return []
        return [header_layout_document]


class AffiliationAddressModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.affiliation_address_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        header_model = document_context.fulltext_models.header_model
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        header_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<header>'
        ).remove_empty_blocks()
        LOGGER.debug('header_layout_document: %r', header_layout_document)
        if not header_layout_document.pages:
            return []
        header_labeled_layout_tokens = get_labeled_layout_tokens_for_model_and_layout_document(
            model=header_model,
            layout_document=header_layout_document,
            document_context=document_context
        )
        semantic_raw_aff_address_list = list(
            SemanticMixedContentWrapper(list(
                header_model.iter_semantic_content_for_labeled_layout_tokens(
                    header_labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawAffiliationAddress)
        )
        LOGGER.info('semantic_raw_aff_address_list count: %d', len(semantic_raw_aff_address_list))
        if not semantic_raw_aff_address_list:
            return []

        return [
            LayoutDocument.for_blocks(
                list(semantic_raw_aff_address.iter_blocks())
            )
            for semantic_raw_aff_address in semantic_raw_aff_address_list
        ]


class NameHeaderModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.name_header_model

    def get_default_tei_sub_directory(
        self,
        tei_training_data_generator: TeiTrainingDataGenerator
    ) -> str:
        return 'name/header/corpus'

    def get_pre_file_path_suffix(self) -> str:
        return '.header'

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        header_model = document_context.fulltext_models.header_model
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        header_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<header>'
        ).remove_empty_blocks()
        LOGGER.debug('header_layout_document: %r', header_layout_document)
        if not header_layout_document.pages:
            return []
        header_labeled_layout_tokens = get_labeled_layout_tokens_for_model_and_layout_document(
            model=header_model,
            layout_document=header_layout_document,
            document_context=document_context
        )
        semantic_raw_author_list = list(
            SemanticMixedContentWrapper(list(
                header_model.iter_semantic_content_for_labeled_layout_tokens(
                    header_labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawAuthors)
        )
        LOGGER.info('semantic_raw_author_list count: %d', len(semantic_raw_author_list))
        if not semantic_raw_author_list:
            return []

        return [
            LayoutDocument.for_blocks([
                block
                for semantic_raw_author in semantic_raw_author_list
                for block in semantic_raw_author.iter_blocks()
            ])
        ]


class NameCitationModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.name_citation_model

    def get_default_tei_sub_directory(
        self,
        tei_training_data_generator: TeiTrainingDataGenerator
    ) -> str:
        return 'name/citation/corpus'

    def get_pre_file_path_suffix(self) -> str:
        return '.citations'

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        reference_segmenter_model = document_context.fulltext_models.reference_segmenter_model
        citation_model = document_context.fulltext_models.citation_model
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        references_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<references>'
        ).remove_empty_blocks()
        reference_segmenter_labeled_layout_tokens = (
            get_labeled_layout_tokens_for_model_and_layout_document(
                model=reference_segmenter_model,
                layout_document=references_layout_document,
                document_context=document_context
            )
        )
        raw_reference_text_list = [
            raw_reference_text
            for raw_reference in SemanticMixedContentWrapper(list(
                reference_segmenter_model.iter_semantic_content_for_labeled_layout_tokens(
                    reference_segmenter_labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawReference)
            for raw_reference_text in raw_reference.iter_by_type(SemanticRawReferenceText)
        ]
        LOGGER.info('raw_reference_text_list count: %d', len(raw_reference_text_list))
        if not raw_reference_text_list:
            return []
        citation_layout_documents = [
            LayoutDocument.for_blocks(
                list(semantic_raw_reference_text.iter_blocks())
            )
            for semantic_raw_reference_text in raw_reference_text_list
        ]
        citation_labeled_layout_tokens_list = (
            get_labeled_layout_tokens_list_for_model_and_layout_documents(
                model=citation_model,
                layout_documents=citation_layout_documents,
                document_context=document_context
            )
        )
        semantic_raw_author_list = [
            raw_author
            for citation_labeled_layout_tokens in citation_labeled_layout_tokens_list
            for raw_author in SemanticMixedContentWrapper(list(
                citation_model.iter_semantic_content_for_labeled_layout_tokens(
                    citation_labeled_layout_tokens
                )
            )).iter_by_type_recursively(SemanticRawAuthors)
        ]
        LOGGER.info('semantic_raw_author_list count: %d', len(semantic_raw_author_list))
        if not semantic_raw_author_list:
            return []

        return [
            LayoutDocument.for_blocks([
                block
                for semantic_raw_author in semantic_raw_author_list
                for block in semantic_raw_author.iter_blocks()
            ])
        ]


class FullTextModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.fulltext_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        body_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<body>'
        ).remove_empty_blocks()
        if not body_layout_document.pages:
            return []
        return [body_layout_document]


class FigureModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.figure_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        fulltext_model = document_context.fulltext_models.fulltext_model
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        body_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<body>'
        ).remove_empty_blocks()
        if not body_layout_document.pages:
            return []
        fulltext_labeled_layout_tokens = get_labeled_layout_tokens_for_model_and_layout_document(
            model=fulltext_model,
            layout_document=body_layout_document,
            document_context=document_context
        )
        raw_figure_list = list(
            SemanticMixedContentWrapper(list(
                fulltext_model.iter_semantic_content_for_labeled_layout_tokens(
                    fulltext_labeled_layout_tokens
                )
            )).iter_by_type_recursively(SemanticRawFigure)
        )
        LOGGER.info('raw_figure_list count: %d', len(raw_figure_list))

        if not raw_figure_list:
            return []
        return [
            LayoutDocument.for_blocks(list(raw_figure.iter_blocks()))
            for raw_figure in raw_figure_list
        ]


class TableModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.table_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        fulltext_model = document_context.fulltext_models.fulltext_model
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        body_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<body>'
        ).remove_empty_blocks()
        if not body_layout_document.pages:
            return []
        fulltext_labeled_layout_tokens = get_labeled_layout_tokens_for_model_and_layout_document(
            model=fulltext_model,
            layout_document=body_layout_document,
            document_context=document_context
        )
        raw_table_list = list(
            SemanticMixedContentWrapper(list(
                fulltext_model.iter_semantic_content_for_labeled_layout_tokens(
                    fulltext_labeled_layout_tokens
                )
            )).iter_by_type_recursively(SemanticRawTable)
        )
        LOGGER.info('raw_table_list count: %d', len(raw_table_list))

        if not raw_table_list:
            return []
        return [
            LayoutDocument.for_blocks(list(raw_table.iter_blocks()))
            for raw_table in raw_table_list
        ]


class ReferenceSegmenterModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.reference_segmenter_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        ref_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<references>'
        ).remove_empty_blocks()
        if not ref_layout_document.pages:
            return []
        return [ref_layout_document]


class CitationModelTrainingDataGenerator(AbstractDocumentModelTrainingDataGenerator):
    def get_main_model(self, document_context: TrainingDataDocumentContext) -> Model:
        return document_context.fulltext_models.citation_model

    def iter_model_layout_documents(
        self,
        layout_document: LayoutDocument,
        document_context: TrainingDataDocumentContext
    ) -> Iterable[LayoutDocument]:
        reference_segmenter_model = document_context.fulltext_models.reference_segmenter_model
        segmentation_label_result = get_segmentation_label_result(
            layout_document,
            document_context=document_context
        )
        references_layout_document = segmentation_label_result.get_filtered_document_by_label(
            '<references>'
        ).remove_empty_blocks()
        reference_segmenter_labeled_layout_tokens = (
            get_labeled_layout_tokens_for_model_and_layout_document(
                model=reference_segmenter_model,
                layout_document=references_layout_document,
                document_context=document_context
            )
        )
        raw_reference_text_list = [
            raw_reference_text
            for raw_reference in SemanticMixedContentWrapper(list(
                reference_segmenter_model.iter_semantic_content_for_labeled_layout_tokens(
                    reference_segmenter_labeled_layout_tokens
                )
            )).iter_by_type(SemanticRawReference)
            for raw_reference_text in raw_reference.iter_by_type(SemanticRawReferenceText)
        ]
        LOGGER.info('raw_reference_text_list count: %d', len(raw_reference_text_list))
        if not raw_reference_text_list:
            return []
        return [
            LayoutDocument.for_blocks(
                list(semantic_raw_reference_text.iter_blocks())
            )
            for semantic_raw_reference_text in raw_reference_text_list
        ]


def generate_training_data_for_layout_document(
    layout_document: LayoutDocument,
    output_path: str,
    source_filename: str,
    document_features_context: DocumentFeaturesContext,
    fulltext_models: FullTextModels,
    use_model: bool,
    use_directory_structure: bool,
    gzip_enabled: bool = False
):
    model_result_cache = ModelResultCache()
    document_context = TrainingDataDocumentContext(
        output_path=output_path,
        source_filename=source_filename,
        document_features_context=document_features_context,
        fulltext_models=fulltext_models,
        use_model=use_model,
        use_directory_structure=use_directory_structure,
        model_result_cache=model_result_cache,
        gzip_enabled=gzip_enabled
    )
    training_data_generators = [
        SegmentationModelTrainingDataGenerator(),
        HeaderModelTrainingDataGenerator(),
        AffiliationAddressModelTrainingDataGenerator(),
        NameHeaderModelTrainingDataGenerator(),
        FullTextModelTrainingDataGenerator(),
        FigureModelTrainingDataGenerator(),
        TableModelTrainingDataGenerator(),
        ReferenceSegmenterModelTrainingDataGenerator(),
        CitationModelTrainingDataGenerator(),
        NameCitationModelTrainingDataGenerator()
    ]
    for training_data_generator in training_data_generators:
        training_data_generator.generate_data_for_layout_document(
            layout_document=layout_document,
            document_context=document_context
        )


def get_layout_document_for_source_filename(
    source_filename: str,
    sciencebeam_parser: ScienceBeamParser,
) -> LayoutDocument:
    with sciencebeam_parser.get_new_session() as session:
        with auto_download_input_file(
            source_filename,
            auto_decompress=True
        ) as local_source_filename:
            source = session.get_source(local_source_filename, MediaTypes.PDF)
            layout_document = source.get_layout_document()
            return layout_document


def generate_training_data_for_source_filename(
    source_filename: str,
    output_path: str,
    sciencebeam_parser: ScienceBeamParser,
    use_model: bool,
    use_directory_structure: bool,
    gzip_enabled: bool
):
    LOGGER.debug('use_model: %r', use_model)
    layout_document = get_layout_document_for_source_filename(
        source_filename,
        sciencebeam_parser=sciencebeam_parser
    )
    generate_training_data_for_layout_document(
        layout_document=layout_document,
        output_path=output_path,
        source_filename=source_filename,
        document_features_context=DocumentFeaturesContext(
            sciencebeam_parser.app_features_context
        ),
        fulltext_models=sciencebeam_parser.fulltext_models,
        use_model=use_model,
        use_directory_structure=use_directory_structure,
        gzip_enabled=gzip_enabled
    )


def get_source_file_list_or_fail(
    source_path_pattern: str
) -> Sequence[str]:
    source_file_list = list(glob(source_path_pattern))
    if not source_file_list:
        raise FileNotFoundError('no files found for file pattern: %r' % source_path_pattern)
    return source_file_list


def run(args: argparse.Namespace):
    LOGGER.info('args: %r', args)
    source_file_list = get_source_file_list_or_fail(args.source_path)
    if args.limit:
        source_file_list = source_file_list[:args.limit]
    LOGGER.info('source files: %d', len(source_file_list))
    output_path = args.output_path
    config = AppConfig.load_yaml(
        DEFAULT_CONFIG_FILE
    )
    sciencebeam_parser = ScienceBeamParser.from_config(config)
    LOGGER.info('output_path: %r', output_path)
    # Note: creating the directory may not be necessary, but provides early feedback
    makedirs(output_path, exist_ok=True)
    for source_filename in source_file_list:
        generate_training_data_for_source_filename(
            source_filename,
            output_path=output_path,
            sciencebeam_parser=sciencebeam_parser,
            use_model=args.use_model,
            use_directory_structure=args.use_directory_structure,
            gzip_enabled=args.gzip
        )


def main(argv: Optional[List[str]] = None):
    LOGGER.debug('argv: %r', argv)
    args = parse_args(argv)
    if args.debug:
        for name in [__name__, 'sciencebeam_parser', 'sciencebeam_trainer_delft']:
            logging.getLogger(name).setLevel('DEBUG')
    run(args)


if __name__ == '__main__':
    logging.basicConfig(level='INFO')

    main()
